from os import makedirs, remove
from uuid import uuid4
import jpype
import jpype.imports
import json
import threading
from pathlib import Path

assert jpype.isJVMStarted

from ca.cgjennings.apps.arkham.project import ProjectUtilities
from ca.cgjennings.apps.arkham.project import Project
from resources import ResourceKit
from java.io import File
import java
from javax.imageio import ImageIO


front_types = {
    "Act.js": "act",
    "ActAssetStory.js": "act",
    "ActEnemy.js": "act",
    "ActLocation.js": "act",
    "ActPortrait.js": "act",
    "Agenda.js": "agenda",
    "AgendaAssetStory.js": "agenda",
    "AgendaEnemy.js": "agenda",
    "AgendaFrontPortrait.js": "agenda",
    "AgendaLocation.js": "agenda",
    "AgendaPortrait.js": "agenda",
    "AgendaTreachery.js": "agenda",
    "Asset.js": "asset",
    "AssetAsset.js": "asset",
    "AssetStory.js": "asset",
    "AssetStoryAsset.js": "asset",
    "AssetStoryEnemy.js": "asset",
    "AssetStoryPortrait.js": "asset",
    "Chaos.js": "chaos",
    "Concealed.js": "concealed",
    "Customizable.js": "customizable",
    "Enemy.js": "enemy",
    "EnemyEnemy.js": "enemy",
    "EnemyLocation.js": "enemy_location",
    "EnemyPortrait.js": "enemy",
    "Event.js": "event",
    "Investigator.js": "investigator",
    "InvestigatorStory.js": "investigator",
    "Key.js": "key",
    "Location.js": "location",
    "LocationLocation.js": "location",
    "Scenario.js": "scenario",
    "Skill.js": "skill",
    "StoryAsset.js": "story",
    "StoryChaos.js": "story",
    "StoryEnemy.js": "story",
    "StoryLocation.js": "story",
    "StoryStory.js": "story",
    "StoryTreachery.js": "story",
    "Treachery.js": "treachery",
    "TreacheryLocation.js": "treachery",
    "TreacheryPortrait.js": "treachery",
    "TreacheryStory.js": "treachery",
    "Ultimatum.js": "ultimatum",
    "WeaknessEnemy.js": "enemy",
    "WeaknessTreachery.js": "treachery",
    "MiniInvestigator.js": "mini",
}

back_types = {
    "Act.js": "act_back",
    "ActAssetStory.js": "asset",
    "ActEnemy.js": "enemy",
    "ActLocation.js": "location",
    "ActPortrait.js": "act_back",
    "Agenda.js": "agenda_back",
    "AgendaAssetStory.js": "asset",
    "AgendaEnemy.js": "enemy",
    "AgendaFrontPortrait.js": "agenda_back",
    "AgendaLocation.js": "location",
    "AgendaPortrait.js": "agenda_back",
    "AgendaTreachery.js": "treachery",
    "Asset.js": "player",
    "AssetAsset.js": "asset",
    "AssetStory.js": "story",
    "AssetStoryAsset.js": "asset",
    "AssetStoryEnemy.js": "enemy",
    "AssetStoryPortrait.js": "story",
    "Chaos.js": "chaos_back",
    "Concealed.js": "concealed_back",
    "Customizable.js": "customizable_back",
    "Enemy.js": "encounter",
    "EnemyEnemy.js": "enemy",
    "EnemyLocation.js": "location_back",
    "EnemyPortrait.js": "encounter",
    "Event.js": "player",
    "Investigator.js": "investigator_back",
    "InvestigatorStory.js": "investigator_back",
    "Key.js": "key_back",
    "Location.js": "location_back",
    "LocationLocation.js": "location",
    "Scenario.js": "scenario_back",
    "Skill.js": "player",
    "StoryAsset.js": "asset",
    "StoryChaos.js": "chaos",
    "StoryEnemy.js": "enemy",
    "StoryLocation.js": "location",
    "StoryStory.js": "story",
    "StoryTreachery.js": "treachery",
    "Treachery.js": "encounter",
    "TreacheryLocation.js": "location",
    "TreacheryPortrait.js": "encounter",
    "TreacheryStory.js": "story",
    "Ultimatum.js": "ultimatum_back",
    "WeaknessEnemy.js": "player",
    "WeaknessTreachery.js": "player",
    "MiniInvestigator.js": "mini_back",
}

PORTRAITS = {
    'Act.js': [ 'Portrait-Front', 'Collection-Both', 'Encounter-Both' ],
    'ActAssetStory.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'ActEnemy.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'ActLocation.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'ActPortrait.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Front', 'Encounter-Front' ],
    'AgendaEnemy.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'AgendaAssetStory.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'AgendaFrontPortrait.js': [ 'Portrait-Front', 'Collection-Both', 'Encounter-Both' ],
    'AgendaLocation.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'AgendaTreachery.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'Asset.js': [ 'Portrait-Front', 'Collection-Front' ],
    'AssetAsset.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Front' ],
    'AssetStoryAsset.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Front', 'Encounter-Front' ],
    'Chaos.js': [ 'Collection-Both', 'Encounter-Both' ],
    'Concealed.js': [ 'Portrait-Front' ],
    'AssetStoryPortrait.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Front', 'Encounter-Front' ],
    'Customizable.js': [ 'Collection-Front' ],
    'AssetStory.js': [ 'Portrait-Front', 'Collection-Front', 'Encounter-Front' ],
    'EnemyEnemy.js': [ 'Portrait-Both', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'MiniInvestigator.js': [ 'Portrait-Both' ],
    'Investigator.js': [ 'TransparentPortrait-Both', 'Portrait-Back', 'Collection-Front' ],
    'Enemy.js': [ 'Portrait-Front', 'Collection-Front', 'Encounter-Front' ],
    'GuideLetter.js': [ 'Portrait1-Front', 'Portrait2-Front' ],
    'EnemyLocation.js': [ 'Portrait-Both', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'EnemyPortrait.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Front', 'Encounter-Front' ],
    'StoryStory.js': [ 'Collection-Both', 'Encounter-Both' ],
    'InvestigatorStory.js': [ 'TransparentPortrait-Both', 'Portrait-Back', 'Collection-Front', 'Encounter-Both' ],
    'TreacheryStory.js': [ 'Portrait-Front', 'Collection-Both', 'Encounter-Both' ],
    'Guide75.js': [ 'Portrait1-Front', 'Portrait2-Front' ],
    'Key.js': [ 'Portrait-Both', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'LocationLocation.js': [ 'Portrait-Both', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'Scenario.js': [ 'Portrait', 'BackPortrait', 'Collection-Both', 'Encounter-Both' ],
    'Agenda.js': [ 'Portrait-Front', 'Collection-Both', 'Encounter-Both' ],
    'AgendaPortrait.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Front', 'Encounter-Front' ],
    'Divider.js': [ 'Encounter-Both' ],
    'TreacheryPortrait.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Front', 'Encounter-Front' ],
    'StoryLocation.js': [ 'BackPortrait-Back', 'Collection-Back', 'Encounter-Both' ],
    'Treachery.js': [ 'Portrait-Front', 'Collection-Front', 'Encounter-Front' ],
    'AssetStoryEnemy.js': [ 'Portrait-Front', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'StoryChaos.js': [ 'Collection-Back', 'Encounter-Both' ],
    'StoryTreachery.js': [ 'BackPortrait-Back', 'Collection-Back', 'Encounter-Both' ],
    'BoxCover.js': [ 'Portrait-Front', 'PortraitBottom-Front' ],
    'Ultimatum.js': [ 'Portrait', 'Collection', 'Encounter' ],
    'StoryAsset.js': [ 'BackPortrait-Back', 'Collection-Back', 'Encounter-Both' ],
    'TreacheryLocation.js': [ 'Portrait-Front', 'Collection-Front', 'Encounter-Front' ],
    'WeaknessTreachery.js': [ 'Portrait-Front', 'Collection-Front', 'Encounter-Front' ],
    'GuideA4.js': [ 'Portrait1-Front', 'Portrait2-Front' ],
    'Event.js': [ 'Portrait-Front', 'Collection-Front', 'Encounter-Both' ],
    'WeaknessEnemy.js': [ 'Portrait-Front', 'Collection-Front', 'Encounter-Front' ],
    'Skill.js': [ 'Portrait-Front', 'Collection-Front', 'Encounter-Both' ],
    'StoryEnemy.js': [ 'BackPortrait-Back', 'Collection-Back', 'Encounter-Both' ],
    'Location.js': [ 'Portrait-Both', 'BackPortrait-Back', 'Collection-Both', 'Encounter-Both' ],
    'PackCover.js': [ 'Portrait-Front' ],
}

def translate_text(value):
    # make sure javastrings are pystrings
    value = str(value)

    value = value.replace('<fullname>', '<name>')
    value = value.replace('<act>', '<action>')
    value = value.replace('<fre>', '<free>')
    value = value.replace('<rea>', '<reaction>')
    value = value.replace('<wil>', '<willpower>')
    value = value.replace('<int>', '<intellect>')
    value = value.replace('<agi>', '<agility>')
    value = value.replace('<com>', '<combat>')
    value = value.replace('<rog>', '<rogue>')
    value = value.replace('<see>', '<seeker>')
    value = value.replace('<sur>', '<survivor>')
    value = value.replace('<gua>', '<guardian>')
    value = value.replace('<mys>', '<mystic>')
    value = value.replace('<sku>', '<skull>')
    value = value.replace('<cul>', '<cultist>')
    value = value.replace('<tab>', '<tablet>')
    value = value.replace('<mon>', '<elder_thing>')
    value = value.replace('<eld>', '<elder_sign>')
    value = value.replace('<ten>', '<fail>')
    value = value.replace('<ble>', '<blessing>')
    value = value.replace('<cur>', '<curse>')
    value = value.replace('<spa>', '<spawn>')
    value = value.replace('<bul>', '<bullet>')
    value = value.replace('<bultab>', '')
    value = value.replace('<vs>\n', '')
    value = value.replace('\n\n', '\n')

    return value

def get_portaits(card):
    script_name = card.getClassName().split("/")[-1]
    bindings = PORTRAITS.get(script_name)
    if not bindings:
        return None
    output = {}
    for index, name in enumerate(bindings):
        output[str(name)] = card.getPortrait(index)
    return output

def extract_images(card, collection, image_folder):
    print(f'extracting images from {card.getName()}')
    script_name = card.getClassName().split("/")[-1]
    bindings = PORTRAITS.get(script_name)
    if not bindings:
        return
    for index, name in enumerate(bindings):
        portrait = card.getPortrait(index)
        source = str(portrait.getSource())
        if source in collection['images']:
            continue
        collection['images'][source] = ''
        portrait_name = source.split('\\')[-1]
        if not portrait_name:
            portrait_name = source.split('/')[-1]
        portrait_format = portrait_name.split('.')[-1]
        new_path = image_folder / portrait_name
        i = 0
        while new_path.exists():
            new_path = image_folder / f'{i}_{portrait_name}'
            i += 1
            if i > 35:
                print(f'{card.getName()} failed to find suitable name for image {portrait_name}')
                continue
        collection['images'][source] = str(new_path)
        outputfile = File(str(new_path))
        ImageIO.write(portrait.getImage(), portrait_format, outputfile)

def determine_encounter_set(card) -> tuple[str|None, str|None]:
    portraits = get_portaits(card)
    if not portraits:
        return None, None
    for title in ('Encounter-Both', 'Encounter-Front', 'Encounter-Back'):
        if title not in portraits:
            continue
        encounter_set_path = str(portraits[title].getSource())
        path = encounter_set_path.split('/')[-1]
        # check for windows path
        if encounter_set_path == path:
            path = path.split('\\')[-1]
        return path, encounter_set_path
    return None, None


class JavaWriter(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, java.lang.String):
            return str(o)
        return super(o)


def run_import(project_path, output_path):
    PROJECT_FOLDER = Path(project_path)
    OUTPUT_FOLDER = Path(output_path)
    OUTPUT_FILE = OUTPUT_FOLDER / 'cards.json'
    IMAGE_FOLDER = OUTPUT_FOLDER / 'images'

    if not OUTPUT_FOLDER.exists():
        makedirs(OUTPUT_FOLDER)
    if not IMAGE_FOLDER.exists():
        makedirs(IMAGE_FOLDER)
    if OUTPUT_FILE.exists():
        remove(OUTPUT_FILE)

    card_files = []
    for root, dir, files in PROJECT_FOLDER.walk():
        card_files += [root/f for f in files if f.split('.')[-1] == 'eon' and f != 'deck.eon']
    print(f'Processing {len(card_files)} cards')

    collection = {
        "name": PROJECT_FOLDER.name,
        "encounter_sets": {},
        "cards": [],
        "images": {},
    }

    step = 0
    while step < len(card_files):
        threads = []
        for file in card_files[step:step+25]:
            thread = threading.Thread(target=parse_card, args=(file, collection, IMAGE_FOLDER))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()
        step += 25

    print(f'Done processing cards. Post processing begins...')

    collection['encounter_sets'] = list(collection['encounter_sets'].values())
    for es in collection['encounter_sets']:
        try:
            es['icon'] = collection['images'][es['icon']]
        except KeyError:
            pass
    del collection['images']

    with open(OUTPUT_FILE, 'w') as file:
        json.dump(collection, file, cls=JavaWriter, indent=4)
    print(f'Done saving.')


def parse_card(file:Path, results:dict, image_folder:Path):
    card = ResourceKit.getGameComponentFromFile(File(str(file)), False)
    if not card:
        print(f'ERROR: {file} appears to have issues loading.')
        return
    script_name = card.getClassName().split("/")[-1]
    if script_name not in front_types and script_name not in back_types:
        return
    settings = card.getSettings()
    out = {}

    extract_images(card, results, image_folder)

    if card.getFullName() != "":
        out["name"] = card.getFullName()
    else:
        out["name"] = file.name


    out["front"] = {
        "type": front_types[script_name],
    }
    out["back"] = {
        "type": back_types[script_name],
    }

    if settings.get('Unique'):
        out["front"]['title'] = "<unique><name>"

    if subtitle := settings.get('Subtitle'):
        out["front"]['subtitle'] = str(subtitle)


    if traits := settings.get("Traits"):
        out['front']['traits'] = traits

    if settings.get("Rules") and settings.get("Keywords"):
        out["front"]["text"] = str(settings.get("Keywords")) + '\n' + str(settings.get("Rules"))
    elif settings.get("Rules") or settings.get("Keywords"):
        out["front"]["text"] = settings.get("Keywords") or settings.get("Rules")


    if health := (settings.get("Health") or settings.get('Stamina')):
        out['front']['health'] = str(health)
    if sanity := settings.get('Sanity'):
        out['front']['sanity'] = str(sanity)
    if evade := settings.get('Evade'):
        out['front']['evade'] = str(evade)
    if combat := settings.get('Attack'):
        out['front']['combat'] = str(combat)

    if damage := settings.get("damage"):
        out["front"]["damage"] = int(damage)
    if horror := settings.get("horror"):
        out["front"]["horror"] = int(horror)

    flavor = (
        settings.get("AgendaStory") or
        settings.get("ActStory") or
        settings.get("Flavor")
    )
    if flavor:
        out["front"]["flavor_text"] = flavor

    if vp := settings.get("Victory"):
        out["front"]["victory"] = vp

    if settings.get("CardClass"):
        out["front"]["classes"] = []
        for c in ('CardClass', 'CardClass2', 'CardClass3'):
            if cl := settings.get(c):
                if cl != None:
                    out['front']['classes'].append(str(cl).lower())

    illustrations = get_portaits(card)
    if illustrations:
        for name, portrait in illustrations.items():
            if name in ('Portrait-Front', 'Portrait-Both', 'TransparentPortrait-Both'):
                out['front']['illustration'] = results['images'][portrait.getSource()]
                # out['front']['illustration_pan_x'] = portrait.getPanX() * 2
                # out['front']['illustration_pan_y'] = portrait.getPanY() * 2
                # out['front']['illustration_scale'] = portrait.getScale() * 2
            if name in ('Portrait-Back', 'Portrait-Both', 'BackPortrait-Back', 'TransparentPortrait-Both'):
                if portrait.getSource() != None:
                    out['back']['illustration'] = results['images'][portrait.getSource()]
                # out['back']['illustration_pan_x'] = portrait.getPanX() * 2
                # out['back']['illustration_pan_y'] = portrait.getPanY() * 2
                # out['back']['illustration_scale'] = portrait.getScale() * 2

    if clues := settings.get("Clues"):
        out['front']['clues'] = clues + ("<per>" if settings.get("PerInvestigator") else '')

    if shroud := settings.get("Shroud"):
        out['front']['shroud'] = shroud + ("<per>" if settings.get("ShroudPerInvestigator") else '')

    if doom := settings.get("Doom"):
        out['front']['doom'] = doom + ("<per>" if settings.get("PerInvestigator") else '')

    if index := settings.get("ScenarioIndex"):
        out['front']['index'] = index + settings.get("ScenarioDeckID")

    if icon := settings.get("LocationIcon"):
        out['front']['connection'] = str(icon).lower()

    if 'Connection1Icon' in settings.getKeySet():
        connections = [
            settings.get("Connection1Icon"),
            settings.get("Connection2Icon"),
            settings.get("Connection3Icon"),
            settings.get("Connection4Icon"),
            settings.get("Connection5Icon"),
            settings.get("Connection6Icon"),
        ]
        out['front']['connections'] = [str(n).lower() for n in connections if n != None]

    if number := settings.get("EncounterNumber"):
        out["encounter_number"] = number

    if number := settings.get("CollectionNumber"):
        out["expansion_number"] = number

    if artist := settings.get("Artist"):
        out['front']['illustrator'] = str(artist)
    if artist := settings.get("ArtistBack"):
        out['back']['illustrator'] = str(artist)

    # add encounter set if needed
    encounter_set, icon = determine_encounter_set(card)
    if encounter_set:
        results['encounter_sets'][encounter_set] = results['encounter_sets'].get(encounter_set, {
            'name': encounter_set,
            'icon': icon,
            'id': str(uuid4()),
            'card_amount': 0,
        })
        out['encounter_set'] = results['encounter_sets'][encounter_set]['id']
        results['encounter_sets'][encounter_set]['card_amount'] += 1

    # field translation
    for field in ('text', 'flavor'):
        for side in (out['front'], out['back']):
            if field in side:
               side[field] = translate_text(side[field])

    # add to result list
    results['cards'].append(out)



"""
Dialogs for creating new cards and projects
"""
from PySide6.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QFormLayout,
    QLineEdit, QPushButton, QComboBox, QLabel,
    QDialogButtonBox, QFileDialog, QMessageBox
)
from PySide6.QtCore import Qt
from pathlib import Path
import json
from uuid import uuid4

import shoggoth
from shoggoth.card import Card, TEMPLATES
from shoggoth.project import Project


class NewCardDialog(QDialog):
    """Dialog for creating a new card"""
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("New Card")
        self.setMinimumWidth(500)
        
        self.selected_template = 'asset'
        self.selected_encounter = None
        
        self.setup_ui()
        
        # Pre-fill from current context if available
        if shoggoth.app.current_card:
            card = shoggoth.app.current_card
            if card.encounter:
                self.selected_encounter = card.encounter
                self.encounter_combo.setCurrentText(card.encounter.name)
                self.template_combo.setCurrentText('treachery')
    
    def setup_ui(self):
        """Setup the UI"""
        layout = QVBoxLayout()
        
        # Form layout for inputs
        form = QFormLayout()
        
        # Card name
        self.name_input = QLineEdit()
        self.name_input.setPlaceholderText("Enter card name")
        form.addRow("Name:", self.name_input)
        
        # Template selection
        self.template_combo = QComboBox()
        templates = [
            'asset', 'event', 'skill', 'investigator',
            'enemy', 'treachery', 'location',
            'act', 'agenda', 'scenario', 'story'
        ]
        self.template_combo.addItems(templates)
        self.template_combo.currentTextChanged.connect(self.on_template_changed)
        form.addRow("Template:", self.template_combo)
        
        # Encounter set selection
        self.encounter_combo = QComboBox()
        self.encounter_combo.addItem("Player Card", None)
        
        if shoggoth.app.current_project:
            for encounter in shoggoth.app.current_project.encounter_sets:
                self.encounter_combo.addItem(encounter.name, encounter)
        
        self.encounter_combo.currentIndexChanged.connect(self.on_encounter_changed)
        form.addRow("Encounter Set:", self.encounter_combo)
        
        layout.addLayout(form)
        
        # Error label
        self.error_label = QLabel()
        self.error_label.setStyleSheet("color: red;")
        layout.addWidget(self.error_label)
        
        # Buttons
        button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        button_box.accepted.connect(self.create_card)
        button_box.rejected.connect(self.reject)
        layout.addWidget(button_box)
        
        self.setLayout(layout)
    
    def on_template_changed(self, template):
        """Handle template change"""
        self.selected_template = template
    
    def on_encounter_changed(self, index):
        """Handle encounter set change"""
        self.selected_encounter = self.encounter_combo.itemData(index)
    
    def create_card(self):
        """Create the card"""
        name = self.name_input.text().strip()
        
        if not name:
            self.error_label.setText("You must enter a card name")
            return
        
        try:
            # Get template
            template = TEMPLATES.get(self.selected_template)
            template['name'] = name
            
            # Create card
            project = shoggoth.app.current_project
            new_card = Card(
                data=template,
                expansion=project,
                encounter=self.selected_encounter
            )
            
            # Add to project
            project.add_card(new_card)
            
            # Refresh tree and navigate to new card
            shoggoth.app.refresh_tree()
            shoggoth.app.show_card(new_card)
            shoggoth.app.select_item_in_tree(new_card.id)
            
            self.accept()
        except Exception as e:
            self.error_label.setText(f"Error creating card: {e}")


class NewProjectDialog(QDialog):
    """Dialog for creating a new project"""
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("New Project")
        self.setMinimumWidth(500)
        
        self.file_path = None
        self.icon_path = None
        
        self.setup_ui()
    
    def setup_ui(self):
        """Setup the UI"""
        layout = QVBoxLayout()
        
        # Form layout
        form = QFormLayout()
        
        # Project code
        self.code_input = QLineEdit()
        self.code_input.setPlaceholderText("e.g., TDL")
        self.code_input.setMaxLength(5)
        form.addRow("Abbreviation:", self.code_input)
        
        # Project name
        self.name_input = QLineEdit()
        self.name_input.setPlaceholderText("e.g., The Dunwich Legacy")
        form.addRow("Name:", self.name_input)
        
        # Icon
        icon_layout = QHBoxLayout()
        self.icon_display = QLineEdit()
        self.icon_display.setReadOnly(True)
        self.icon_display.setPlaceholderText("No icon selected")
        icon_layout.addWidget(self.icon_display)
        
        icon_btn = QPushButton("Browse")
        icon_btn.clicked.connect(self.browse_icon)
        icon_layout.addWidget(icon_btn)
        
        form.addRow("Icon:", icon_layout)
        
        # File location
        file_layout = QHBoxLayout()
        self.file_display = QLineEdit()
        self.file_display.setReadOnly(True)
        self.file_display.setPlaceholderText("No location selected")
        file_layout.addWidget(self.file_display)
        
        file_btn = QPushButton("Browse")
        file_btn.clicked.connect(self.browse_file)
        file_layout.addWidget(file_btn)
        
        form.addRow("Save Location:", file_layout)
        
        layout.addLayout(form)
        
        # Error label
        self.error_label = QLabel()
        self.error_label.setStyleSheet("color: red;")
        layout.addWidget(self.error_label)
        
        # Buttons
        button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        button_box.accepted.connect(self.create_project)
        button_box.rejected.connect(self.reject)
        layout.addWidget(button_box)
        
        self.setLayout(layout)
    
    def browse_icon(self):
        """Browse for icon file"""
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Select Icon",
            str(Path.home()),
            "Images (*.png *.jpg *.jpeg *.webp)"
        )
        if file_path:
            self.icon_path = file_path
            self.icon_display.setText(file_path)
    
    def browse_file(self):
        """Browse for save location"""
        suggested_name = self.name_input.text() or "project"
        file_path, _ = QFileDialog.getSaveFileName(
            self,
            "Save Project As",
            str(Path.home() / f"{suggested_name}.json"),
            "JSON Files (*.json)"
        )
        if file_path:
            self.file_path = file_path
            self.file_display.setText(file_path)
    
    def create_project(self):
        """Create the project"""
        name = self.name_input.text().strip()
        code = self.code_input.text().strip()
        
        if not name:
            self.error_label.setText("You must enter a project name")
            return
        
        if not code:
            self.error_label.setText("You must enter an abbreviation")
            return
        
        if not self.file_path:
            self.error_label.setText("You must select a save location")
            return
        
        try:
            # Create project data
            data = Project.new(
                name=name,
                code=code,
                icon=self.icon_path or ''
            )
            
            # Save to file
            with open(self.file_path, 'w') as f:
                json.dump(data, f, indent=4)
            
            # Load project
            shoggoth.app.open_project(self.file_path)
            
            self.accept()
        except Exception as e:
            self.error_label.setText(f"Error creating project: {e}")


"""
Editor widgets for Shoggoth using PySide6
"""
import json
from PySide6.QtWidgets import (
    QWidget, QVBoxLayout, QHBoxLayout, QFormLayout,
    QLineEdit, QTextEdit, QPushButton, QLabel, QComboBox,
    QScrollArea, QGroupBox, QSpinBox, QCompleter
)
from PySide6.QtCore import Qt, Signal, QSize, QPropertyAnimation, QEasingCurve
from PySide6.QtGui import QPixmap, QIcon

from shoggoth.text_editor import ArkhamTextEdit, ArkhamTextHighlighter
from shoggoth.files import overlay_dir
from shoggoth.floating_label_widget import FloatingLabelTextEdit, FloatingLabelLineEdit
from shoggoth import face_editors

# Known traits for autocomplete (loaded from highlighter)
KNOWN_TRAITS = sorted(ArkhamTextHighlighter(None).known_traits)


class FieldWidget:
    """Base class for field widgets that sync with card data"""

    def __init__(self, widget, card_key, converter=str, deconverter=str):
        self.widget = widget
        self.card_key = card_key
        self.converter = converter
        self.deconverter = deconverter
        self._updating = False

    def update_from_card(self, card_data):
        """Update widget from card data"""
        self._updating = True
        value = card_data.data.get(self.card_key)

        if value == '<copy>':
            self.set_widget_value(value)
        else:
            self.set_widget_value(self.deconverter(value) if value else '')
        self._updating = False

    def update_card(self, card_data, value):
        """Update card from widget value"""
        if self._updating:
            return False

        try:
            if value == '<copy>':
                card_data.set(self.card_key, value)
            else:
                card_data.set(self.card_key, self.converter(value) if value else None)
            return True
        except ValueError as e:
            print(f'Error updating card: {e}')
            return False

    def set_widget_value(self, value):
        """Set the widget's value - override in subclasses"""
        if isinstance(self.widget, QLineEdit):
            self.widget.setText(str(value))
        elif isinstance(self.widget, QTextEdit):
            self.widget.setPlainText(str(value))
        elif isinstance(self.widget, QComboBox):
            # For editable comboboxes, set the text directly
            if self.widget.isEditable():
                self.widget.setCurrentText(str(value))
            else:
                self.widget.setCurrentText(str(value))

    def get_widget_value(self):
        """Get the widget's current value"""
        if isinstance(self.widget, QLineEdit):
            return self.widget.text()
        elif isinstance(self.widget, QTextEdit):
            return self.widget.toPlainText()
        elif isinstance(self.widget, QComboBox):
            return self.widget.currentText()
        return ''


class LabeledLineEdit(QWidget):
    """A labeled line edit widget with floating label"""

    textChanged = Signal(str)

    def __init__(self, label_text):
        super().__init__()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.floating_widget = FloatingLabelLineEdit(label_text)
        self.input = self.floating_widget.input
        self.input.textChanged.connect(self.textChanged.emit)

        layout.addWidget(self.floating_widget)
        self.setLayout(layout)

    def text(self):
        return self.input.text()

    def setText(self, text):
        self.floating_widget.setText(text)


class TraitLineEdit(QLineEdit):
    """Line edit with autocomplete for Arkham Horror traits.

    Supports multiple traits separated by ". " (e.g., "Humanoid. Creature. Monster.")
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        # Create completer with known traits
        self.completer = QCompleter(KNOWN_TRAITS)
        self.completer.setWidget(self)
        self.completer.setCompletionMode(QCompleter.PopupCompletion)
        self.completer.setCaseSensitivity(Qt.CaseInsensitive)
        self.completer.setFilterMode(Qt.MatchContains)
        self.completer.activated.connect(self.insert_completion)

        # Connect text changes to update completer
        self.textChanged.connect(self.update_completer)

    def get_current_word_bounds(self):
        """Get the start and end positions of the current word being typed."""
        text = self.text()
        cursor_pos = self.cursorPosition()

        # Find start: look for ". " or start of string
        start = 0
        search_pos = cursor_pos - 1
        while search_pos >= 0:
            if search_pos >= 1 and text[search_pos-1:search_pos+1] == '. ':
                start = search_pos + 1
                break
            search_pos -= 1

        # Skip leading whitespace
        while start < cursor_pos and text[start] == ' ':
            start += 1

        return start, cursor_pos

    def update_completer(self):
        """Update completer prefix based on current word being typed."""
        start, end = self.get_current_word_bounds()
        current_word = self.text()[start:end].strip()

        if current_word and len(current_word) >= 1:
            self.completer.setCompletionPrefix(current_word)
            if self.completer.completionCount() > 0:
                self.completer.complete()
            else:
                self.completer.popup().hide()
        else:
            self.completer.popup().hide()

    def insert_completion(self, completion):
        """Insert the selected completion with proper formatting."""
        text = self.text()
        start, end = self.get_current_word_bounds()

        # Format trait: capitalize first letter, add dot
        formatted_trait = completion[0].upper() + completion[1:] + '.'

        # Build new text
        before = text[:start]
        after = text[end:]

        new_text = before + formatted_trait + after
        self.setText(new_text)

        # Position cursor after the inserted trait
        new_cursor_pos = start + len(formatted_trait)
        self.setCursorPosition(new_cursor_pos)

    def keyPressEvent(self, event):
        """Handle key press events for completer interaction."""
        if self.completer.popup().isVisible():
            if event.key() in (Qt.Key_Enter, Qt.Key_Return):
                if self.completer.currentCompletion():
                    self.insert_completion(self.completer.currentCompletion())
                    self.completer.popup().hide()
                    return

        super().keyPressEvent(event)


class LabeledTraitEdit(QWidget):
    """A labeled line edit widget with trait autocomplete and floating label."""

    textChanged = Signal(str)

    def __init__(self, label_text="Traits"):
        super().__init__()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        # Use FloatingLabelLineEdit but replace the input with TraitLineEdit
        self.floating_widget = FloatingLabelLineEdit(label_text)

        # Replace the standard QLineEdit with TraitLineEdit
        old_input = self.floating_widget.input
        self.floating_widget.input = TraitLineEdit()
        self.floating_widget.input.setStyleSheet(old_input.styleSheet())

        # Replace in layout
        self.floating_widget.layout().removeWidget(old_input)
        old_input.deleteLater()
        self.floating_widget.layout().addWidget(self.floating_widget.input)

        # Reconnect events
        self.floating_widget.input.textChanged.connect(self.floating_widget.on_text_changed)
        self.floating_widget.input.installEventFilter(self.floating_widget)

        self.input = self.floating_widget.input
        self.input.textChanged.connect(self.textChanged.emit)

        layout.addWidget(self.floating_widget)
        self.setLayout(layout)

    def text(self):
        return self.input.text()

    def setText(self, text):
        self.floating_widget.setText(text)


class LabeledTextEdit(QWidget):
    """A labeled text edit widget with floating label"""

    textChanged = Signal()

    def __init__(self, label_text, use_arkham_editor=False):
        super().__init__()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        # For Arkham text editor, we need to replace the input after creation
        self.floating_widget = FloatingLabelTextEdit(label_text)

        if use_arkham_editor:
            # Replace the standard QTextEdit with ArkhamTextEdit
            old_input = self.floating_widget.input
            self.floating_widget.input = ArkhamTextEdit()
            self.floating_widget.input.setStyleSheet(old_input.styleSheet())
            self.floating_widget.input.setMinimumHeight(100)
            self.floating_widget.input.setMaximumHeight(140)

            # Replace in layout
            self.floating_widget.layout().removeWidget(old_input)
            old_input.deleteLater()
            self.floating_widget.layout().addWidget(self.floating_widget.input)

            # Reconnect events
            self.floating_widget.input.textChanged.connect(self.floating_widget.on_text_changed)
            self.floating_widget.input.installEventFilter(self.floating_widget)

        self.input = self.floating_widget.input
        self.input.textChanged.connect(self.textChanged.emit)

        layout.addWidget(self.floating_widget)
        self.setLayout(layout)

    def toPlainText(self):
        return self.input.toPlainText()

    def setPlainText(self, text):
        self.floating_widget.setPlainText(text)


class IconsWidget(QWidget):
    """Widget for skill icons with +/- buttons for each icon type (WICAQ)"""

    # Icon order and their full names
    ICONS = [
        ('W', 'Willpower'),
        ('I', 'Intellect'),
        ('C', 'Combat'),
        ('A', 'Agility'),
        ('Q', 'Wild'),
    ]
    MAX_COUNT = 8

    iconsChanged = Signal(str)

    def __init__(self):
        super().__init__()
        self.counts = {letter: 0 for letter, _ in self.ICONS}
        self.labels = {}
        self._updating = False

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(12)

        # Label
        icons_label = QLabel("Icons")
        icons_label.setMinimumWidth(50)
        layout.addWidget(icons_label)

        # Create +/- controls for each icon
        for letter, name in self.ICONS:
            icon_layout = QHBoxLayout()
            icon_layout.setSpacing(2)

            # Icon image
            icon_label = QLabel()
            icon_label.setToolTip(name)
            icon_path = overlay_dir / f"skill_icon_{letter}.png"
            if icon_path.exists():
                pixmap = QPixmap(str(icon_path)).scaled(
                    20, 20, Qt.KeepAspectRatio, Qt.SmoothTransformation
                )
                icon_label.setPixmap(pixmap)
            else:
                icon_label.setText(letter)
            icon_label.setFixedSize(20, 20)
            icon_layout.addWidget(icon_label)

            # Minus button
            minus_btn = QPushButton("-")
            minus_btn.setFixedSize(24, 24)
            minus_btn.clicked.connect(lambda checked, l=letter: self.decrement(l))
            icon_layout.addWidget(minus_btn)

            # Count label
            count_label = QLabel("0")
            count_label.setAlignment(Qt.AlignCenter)
            count_label.setFixedWidth(20)
            self.labels[letter] = count_label
            icon_layout.addWidget(count_label)

            # Plus button
            plus_btn = QPushButton("+")
            plus_btn.setFixedSize(24, 24)
            plus_btn.clicked.connect(lambda checked, l=letter: self.increment(l))
            icon_layout.addWidget(plus_btn)

            layout.addLayout(icon_layout)

        layout.addStretch()
        self.setLayout(layout)

    def increment(self, letter):
        """Increment count for an icon"""
        if self.counts[letter] < self.MAX_COUNT:
            self.counts[letter] += 1
            self.labels[letter].setText(str(self.counts[letter]))
            if not self._updating:
                self.iconsChanged.emit(self.get_icons_string())

    def decrement(self, letter):
        """Decrement count for an icon"""
        if self.counts[letter] > 0:
            self.counts[letter] -= 1
            self.labels[letter].setText(str(self.counts[letter]))
            if not self._updating:
                self.iconsChanged.emit(self.get_icons_string())

    def get_icons_string(self):
        """Get the icons as an ordered string like 'WWICQQ'"""
        result = ''
        for letter, _ in self.ICONS:
            result += letter * self.counts[letter]
        return result

    def set_icons_string(self, icons_str):
        """Set icons from a string like 'WWICQQ'"""
        self._updating = True
        # Reset all counts
        for letter, _ in self.ICONS:
            self.counts[letter] = 0

        # Count occurrences of each letter
        if icons_str:
            for char in icons_str.upper():
                if char in self.counts:
                    self.counts[char] = min(self.counts[char] + 1, self.MAX_COUNT)

        # Update labels
        for letter, _ in self.ICONS:
            self.labels[letter].setText(str(self.counts[letter]))

        self._updating = False


class IllustrationWidget(QWidget):
    """Widget for illustration settings"""

    # Signal emitted when illustration mode is toggled (enabled, face_side)
    illustration_mode_changed = Signal(bool, str)

    def __init__(self, face_side='front'):
        super().__init__()
        self.face_side = face_side
        self.illustration_mode = False
        layout = QVBoxLayout()

        # Image path
        path_layout = QHBoxLayout()
        self.path_input = LabeledLineEdit("Image Path")
        path_layout.addWidget(self.path_input)
        browse_btn = QPushButton("Browse")
        browse_btn.clicked.connect(self.browse_image)
        path_layout.addWidget(browse_btn)
        layout.addLayout(path_layout)

        # Pan and scale with edit button
        pan_scale_layout = QHBoxLayout()
        self.pan_y_input = LabeledLineEdit("Pan Y")
        self.pan_x_input = LabeledLineEdit("Pan X")
        self.scale_input = LabeledLineEdit("Scale")
        pan_scale_layout.addWidget(self.pan_y_input)
        pan_scale_layout.addWidget(self.pan_x_input)
        pan_scale_layout.addWidget(self.scale_input)

        # Edit position button
        self.edit_position_btn = QPushButton("Edit Position")
        self.edit_position_btn.setCheckable(True)
        self.edit_position_btn.setToolTip("Drag on preview to position image, scroll to scale")
        self.edit_position_btn.clicked.connect(self.toggle_illustration_mode)
        pan_scale_layout.addWidget(self.edit_position_btn)

        layout.addLayout(pan_scale_layout)

        # Artist
        self.artist_input = LabeledLineEdit("Artist")
        layout.addWidget(self.artist_input)

        self.setLayout(layout)

    def toggle_illustration_mode(self, checked):
        """Toggle illustration positioning mode"""
        self.illustration_mode = checked
        if checked:
            self.edit_position_btn.setText("Done")
            self.edit_position_btn.setStyleSheet("background-color: #4a9eff; color: white;")
        else:
            self.edit_position_btn.setText("Edit Position")
            self.edit_position_btn.setStyleSheet("")
        self.illustration_mode_changed.emit(checked, self.face_side)

    def set_illustration_mode(self, enabled):
        """Set illustration mode programmatically"""
        self.edit_position_btn.setChecked(enabled)
        self.toggle_illustration_mode(enabled)

    def update_pan(self, delta_x, delta_y):
        """Update pan values from external input (e.g., preview drag)"""
        try:
            current_x = int(self.pan_x_input.text() or 0)
            current_y = int(self.pan_y_input.text() or 0)
            self.pan_x_input.setText(str(current_x + delta_x))
            self.pan_y_input.setText(str(current_y + delta_y))
        except ValueError:
            pass

    def update_scale(self, delta):
        """Update scale value from external input (e.g., preview scroll)"""
        try:
            current_scale = float(self.scale_input.text() or 1.0)
            new_scale = max(0.1, current_scale + delta)
            self.scale_input.setText(f"{new_scale:.3f}")
        except ValueError:
            pass

    def browse_image(self):
        """Browse for image file"""
        from PySide6.QtWidgets import QFileDialog
        from pathlib import Path

        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Select Image",
            str(Path.home()),
            "Images (*.png *.jpg *.jpeg *.webp)"
        )
        if file_path:
            self.path_input.setText(file_path)


class CardEditor(QWidget):
    """Main card editor widget"""

    # Signal emitted when card data changes
    data_changed = Signal()

    def __init__(self, card):
        super().__init__()
        self.card = card

        # Main layout with scroll
        scroll = QScrollArea()
        scroll.setWidgetResizable(True)

        content = QWidget()
        layout = QVBoxLayout()

        # Basic info section (collapsible)
        basic_group = QWidget()
        basic_layout = QVBoxLayout()
        basic_layout.setContentsMargins(0, 0, 0, 0)
        basic_layout.setSpacing(4)

        # Header row with Name field and toggle button on the right
        header_row = QHBoxLayout()
        header_row.setContentsMargins(0, 0, 0, 0)

        self.name_input = LabeledLineEdit("Name")
        header_row.addWidget(self.name_input)

        self.basic_toggle_btn = QPushButton("\u25b6")  # Right arrow
        self.basic_toggle_btn.setFixedWidth(28)
        self.basic_toggle_btn.setCheckable(True)
        self.basic_toggle_btn.setChecked(False)
        self.basic_toggle_btn.setToolTip("Show/hide basic card info")
        self.basic_toggle_btn.clicked.connect(self.toggle_basic_info)
        header_row.addWidget(self.basic_toggle_btn)

        basic_layout.addLayout(header_row)

        # Collapsible content (hidden by default)
        self.basic_info_content = QWidget()
        content_layout = QVBoxLayout()
        content_layout.setContentsMargins(0, 0, 0, 0)

        self.copyright_input = LabeledLineEdit("Copyright")
        content_layout.addWidget(self.copyright_input)

        # Numbers
        numbers_layout = QHBoxLayout()
        self.amount_input = LabeledLineEdit("Amount in set")
        self.collection_input = LabeledLineEdit("Collection #")
        self.encounter_input = LabeledLineEdit("Encounter Set #")
        numbers_layout.addWidget(self.amount_input)
        numbers_layout.addWidget(self.collection_input)
        numbers_layout.addWidget(self.encounter_input)
        content_layout.addLayout(numbers_layout)

        self.investigator_input = LabeledLineEdit("Investigator Link")
        content_layout.addWidget(self.investigator_input)

        self.id_input = LabeledLineEdit("ID")
        self.id_input.input.setReadOnly(True)
        content_layout.addWidget(self.id_input)

        self.basic_info_content.setLayout(content_layout)
        self.basic_info_content.setMaximumHeight(0)
        basic_layout.addWidget(self.basic_info_content)

        # Animation for collapsible content
        self.basic_info_animation = QPropertyAnimation(self.basic_info_content, b"maximumHeight")
        self.basic_info_animation.setDuration(150)
        self.basic_info_animation.setEasingCurve(QEasingCurve.OutCubic)

        basic_group.setLayout(basic_layout)
        layout.addWidget(basic_group)

        # Add JSON view toggle at card level
        json_button_row = QHBoxLayout()
        self.json_view_btn = QPushButton("📝 View Card as JSON")
        self.json_view_btn.setMaximumWidth(200)
        self.json_view_btn.clicked.connect(self.toggle_json_view)
        json_button_row.addWidget(self.json_view_btn)
        json_button_row.addStretch()
        json_widget = QWidget()
        json_widget.setLayout(json_button_row)
        layout.addWidget(json_widget)

        # Container for face editors or JSON editor
        self.editor_container = QWidget()
        self.editor_layout = QVBoxLayout()
        self.editor_container.setLayout(self.editor_layout)
        layout.addWidget(self.editor_container)

        # Track whether showing JSON
        self.showing_json = False

        # Create initial editors (form view)
        self.create_form_editors()

        content.setLayout(layout)
        scroll.setWidget(content)

        # Main widget layout
        main_layout = QVBoxLayout()
        main_layout.addWidget(scroll)
        self.setLayout(main_layout)

        # Setup card field bindings
        self.fields = []
        self.setup_card_fields()
        self.load_card()

    def create_form_editors(self):
        """Create the form-based face editors"""
        # Clear container
        while self.editor_layout.count():
            item = self.editor_layout.takeAt(0)
            if item.widget():
                item.widget().deleteLater()

        # Front face editor
        front_group = QGroupBox("Front")
        self.front_layout = QVBoxLayout()
        self.front_editor = None
        self.create_front_editor()
        front_group.setLayout(self.front_layout)
        self.editor_layout.addWidget(front_group)

        # Back face editor
        back_group = QGroupBox("Back")
        self.back_layout = QVBoxLayout()
        self.back_editor = None
        self.create_back_editor()
        back_group.setLayout(self.back_layout)
        self.editor_layout.addWidget(back_group)

    def create_json_editor(self):
        """Create the JSON editor for entire card"""
        # Clear container
        while self.editor_layout.count():
            item = self.editor_layout.takeAt(0)
            if item.widget():
                item.widget().deleteLater()

        # Create JSON editor group
        json_group = QGroupBox("Card JSON")
        json_layout = QVBoxLayout()

        # Info
        info = QLabel("Edit the entire card data as JSON. Changes update automatically when valid.")
        info.setWordWrap(True)
        info.setStyleSheet("color: #666; padding: 5px;")
        json_layout.addWidget(info)

        # JSON editor
        from shoggoth.editors import ArkhamTextEdit
        self.json_editor = ArkhamTextEdit()

        # Enable JSON syntax highlighting
        try:
            from pygments.lexers import JsonLexer
            self.json_editor.set_lexer(JsonLexer())
        except:
            pass

        self.json_editor.textChanged.connect(self.on_json_changed)
        json_layout.addWidget(self.json_editor)

        # Status
        self.json_status = QLabel("")
        json_layout.addWidget(self.json_status)

        # Buttons
        button_row = QHBoxLayout()
        format_btn = QPushButton("Format JSON")
        format_btn.clicked.connect(self.format_json)
        button_row.addWidget(format_btn)

        validate_btn = QPushButton("Validate")
        validate_btn.clicked.connect(self.validate_json)
        button_row.addWidget(validate_btn)

        button_row.addStretch()
        json_layout.addLayout(button_row)

        json_group.setLayout(json_layout)
        self.editor_layout.addWidget(json_group)

        # Load current card data
        self.load_json_data()

    def toggle_json_view(self):
        """Toggle between form editors and JSON editor"""
        self.showing_json = not self.showing_json

        if self.showing_json:
            self.json_view_btn.setText("View as Form")
            self.create_json_editor()
        else:
            self.json_view_btn.setText("View Card as JSON")
            self.create_form_editors()

    def toggle_basic_info(self, checked):
        """Toggle visibility of basic info fields with animation"""
        # Update arrow icon
        self.basic_toggle_btn.setText("\u25bc" if checked else "\u25b6")  # Down / Right arrow

        # Animate the content height
        self.basic_info_animation.stop()
        if checked:
            # Expanding: animate from 0 to full height
            target_height = self.basic_info_content.sizeHint().height()
            self.basic_info_animation.setStartValue(0)
            self.basic_info_animation.setEndValue(target_height)
        else:
            # Collapsing: animate from current height to 0
            self.basic_info_animation.setStartValue(self.basic_info_content.height())
            self.basic_info_animation.setEndValue(0)
        self.basic_info_animation.start()

    def load_json_data(self):
        """Load card data into JSON editor"""
        try:
            # Get entire card data
            card_data = {
                'name': self.card.name,
                'id': self.card.id,
                'amount': self.card.amount,
                'expansion_number': self.card.expansion_number,
                'encounter_number': self.card.encounter_number,
                'investigator': self.card.get('investigator'),
                'front': self.card.front.data,
                'back': self.card.back.data,
            }
            # Remove None values
            card_data = {k: v for k, v in card_data.items() if v is not None}

            json_text = json.dumps(card_data, indent=2)
            self.json_editor.setPlainText(json_text)
            self.json_status.setText("✓ JSON loaded")
            self.json_status.setStyleSheet("color: green; padding: 5px;")
        except Exception as e:
            self.json_status.setText(f"✗ Load error: {e}")
            self.json_status.setStyleSheet("color: red; padding: 5px;")

    def on_json_changed(self):
        """Handle JSON editor changes"""
        if not hasattr(self, 'json_editor'):
            return

        try:
            text = self.json_editor.toPlainText()
            if not text.strip():
                return

            data = json.loads(text)

            # Update card data
            if 'name' in data:
                self.card.set('name', data['name'])
            if 'amount' in data:
                self.card.set('amount', data['amount'])
            if 'expansion_number' in data:
                self.card.expansion_number = data['expansion_number']
            if 'encounter_number' in data:
                self.card.encounter_number = data['encounter_number']
            if 'investigator' in data:
                self.card.set('investigator', data['investigator'])

            # Update face data
            if 'front' in data:
                self.card.front.data = data['front']
            if 'back' in data:
                self.card.back.data = data['back']

            # Emit signal for preview update
            self.data_changed.emit()

            self.json_status.setText("✓ Saved")
            self.json_status.setStyleSheet("color: green; padding: 5px;")

        except json.JSONDecodeError as e:
            self.json_status.setText(f"✗ Invalid JSON: {str(e)[:50]}...")
            self.json_status.setStyleSheet("color: red; padding: 5px;")

    def format_json(self):
        """Format the JSON"""
        try:
            text = self.json_editor.toPlainText()
            if text.strip():
                data = json.loads(text)
                formatted = json.dumps(data, indent=2)
                self.json_editor.setPlainText(formatted)
                self.json_status.setText("✓ Formatted")
                self.json_status.setStyleSheet("color: green; padding: 5px;")
        except json.JSONDecodeError as e:
            self.json_status.setText(f"✗ Invalid JSON: {e}")
            self.json_status.setStyleSheet("color: red; padding: 5px;")

    def validate_json(self):
        """Validate the JSON"""
        try:
            text = self.json_editor.toPlainText()
            if text.strip():
                json.loads(text)
                self.json_status.setText("✓ Valid JSON")
                self.json_status.setStyleSheet("color: green; padding: 5px;")
            else:
                self.json_status.setText("⚠ Empty")
                self.json_status.setStyleSheet("color: orange; padding: 5px;")
        except json.JSONDecodeError as e:
            self.json_status.setText(f"✗ Invalid: {e}")
            self.json_status.setStyleSheet("color: red; padding: 5px;")

    def create_front_editor(self):
        """Create/recreate front face editor"""
        # Remove old editor if exists
        if self.front_editor:
            self.front_layout.removeWidget(self.front_editor)
            self.front_editor.setParent(None)
            self.front_editor.deleteLater()

        # Create new editor
        self.front_editor = face_editors.get_editor_for_face(self.card.front)
        self.front_editor.type_changed.connect(lambda: self.on_type_changed('front'))
        self.front_layout.addWidget(self.front_editor)

    def create_back_editor(self):
        """Create/recreate back face editor"""
        # Remove old editor if exists
        if self.back_editor:
            self.back_layout.removeWidget(self.back_editor)
            self.back_editor.setParent(None)
            self.back_editor.deleteLater()

        # Create new editor
        self.back_editor = face_editors.get_editor_for_face(self.card.back)
        self.back_editor.type_changed.connect(lambda: self.on_type_changed('back'))
        self.back_layout.addWidget(self.back_editor)

    def on_type_changed(self, which_face):
        """Handle face type change - recreate the appropriate editor"""
        if which_face == 'front':
            self.create_front_editor()
        else:
            self.create_back_editor()

    def setup_card_fields(self):
        """Setup bindings for card-level fields"""
        self.fields = [
            FieldWidget(self.name_input.input, 'name'),
            FieldWidget(self.copyright_input.input, 'copyright'),
            FieldWidget(self.amount_input.input, 'amount', int),
            FieldWidget(self.collection_input.input, 'expansion_number'),
            FieldWidget(self.encounter_input.input, 'encounter_number'),
            FieldWidget(self.investigator_input.input, 'investigator'),
            FieldWidget(self.id_input.input, 'id'),
        ]

        # Connect signals
        for field in self.fields:
            widget = field.widget
            if isinstance(widget, QLineEdit):
                widget.textChanged.connect(lambda v, f=field: self.on_field_changed(f, v))

    def load_card(self):
        """Load card data into fields"""
        for field in self.fields:
            field.update_from_card(self.card)

    def on_field_changed(self, field, value):
        """Handle field value changes"""
        if field.update_card(self.card, value):
            self.data_changed.emit()

    def cleanup(self):
        """Cleanup editor resources"""
        if self.front_editor and hasattr(self.front_editor, 'cleanup'):
            self.front_editor.cleanup()
        if self.back_editor and hasattr(self.back_editor, 'cleanup'):
            self.back_editor.cleanup()


"""
Project editor widget for Shoggoth using PySide6
"""
from PySide6.QtWidgets import (
    QWidget, QVBoxLayout, QHBoxLayout, QFormLayout,
    QLineEdit, QPushButton, QLabel, QFileDialog,
    QScrollArea, QGridLayout, QGroupBox, QInputDialog,
    QMessageBox
)
from PySide6.QtCore import Qt, Signal, Slot
from PySide6.QtGui import QPixmap, QImage
from pathlib import Path
import threading

import shoggoth


class ProjectEditor(QWidget):
    """Editor widget for project settings"""

    # Signal emitted when project data changes
    data_changed = Signal()

    # Signal for adding thumbnails from background thread
    thumbnail_ready = Signal(QPixmap, str, int)

    def __init__(self, project, card_renderer=None):
        super().__init__()
        self.project = project
        self.card_renderer = card_renderer
        self._updating = False

        # Connect thumbnail signal
        self.thumbnail_ready.connect(self._add_thumbnail)

        self.setup_ui()
        self.load_data()

        # Load thumbnails in background
        if self.card_renderer:
            threading.Thread(target=self.load_thumbnails, daemon=True).start()

    def setup_ui(self):
        """Setup the user interface"""
        layout = QVBoxLayout()

        # Project info section
        info_group = QGroupBox("Project Information")
        info_layout = QFormLayout()

        # Name
        self.name_input = QLineEdit()
        self.name_input.textChanged.connect(lambda: self.on_field_changed('name'))
        info_layout.addRow("Name:", self.name_input)

        # Code
        self.code_input = QLineEdit()
        self.code_input.setPlaceholderText("Short code for file naming")
        self.code_input.textChanged.connect(lambda: self.on_field_changed('code'))
        info_layout.addRow("Code:", self.code_input)

        # Default copyright
        self.copyright_input = QLineEdit()
        self.copyright_input.setPlaceholderText("Default copyright text for cards")
        self.copyright_input.textChanged.connect(lambda: self.on_field_changed('default_copyright'))
        info_layout.addRow("Default Copyright:", self.copyright_input)

        # Icon
        icon_layout = QHBoxLayout()
        self.icon_input = QLineEdit()
        self.icon_input.textChanged.connect(lambda: self.on_field_changed('icon'))
        icon_layout.addWidget(self.icon_input)
        browse_btn = QPushButton("Browse")
        browse_btn.clicked.connect(self.browse_icon)
        icon_layout.addWidget(browse_btn)
        info_layout.addRow("Icon:", icon_layout)

        info_group.setLayout(info_layout)
        layout.addWidget(info_group)

        # Actions section
        actions_group = QGroupBox("Actions")
        actions_layout = QHBoxLayout()

        add_encounter_btn = QPushButton("Add Encounter Set")
        add_encounter_btn.clicked.connect(self.add_encounter_set)
        actions_layout.addWidget(add_encounter_btn)

        actions_layout.addStretch()

        actions_group.setLayout(actions_layout)
        layout.addWidget(actions_group)

        # Statistics section
        stats_group = QGroupBox("Statistics")
        stats_layout = QFormLayout()

        self.encounter_count_label = QLabel("0")
        stats_layout.addRow("Encounter Sets:", self.encounter_count_label)

        self.card_count_label = QLabel("0")
        stats_layout.addRow("Total Cards:", self.card_count_label)

        self.player_card_count_label = QLabel("0")
        stats_layout.addRow("Player Cards:", self.player_card_count_label)

        stats_group.setLayout(stats_layout)
        layout.addWidget(stats_group)

        # Card thumbnails section
        thumbnails_group = QGroupBox("Card Overview")
        thumbnails_layout = QVBoxLayout()

        # Scroll area for thumbnails
        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setMinimumHeight(300)

        self.thumbnail_container = QWidget()
        self.thumbnail_grid = QGridLayout()
        self.thumbnail_grid.setSpacing(10)
        self.thumbnail_container.setLayout(self.thumbnail_grid)

        scroll.setWidget(self.thumbnail_container)
        thumbnails_layout.addWidget(scroll)

        thumbnails_group.setLayout(thumbnails_layout)
        layout.addWidget(thumbnails_group)

        self.setLayout(layout)

    def load_data(self):
        """Load project data into fields"""
        self._updating = True

        self.name_input.setText(self.project.get('name', ''))
        self.code_input.setText(self.project.get('code', ''))
        self.copyright_input.setText(self.project.get('default_copyright', ''))
        self.icon_input.setText(self.project.get('icon', ''))

        # Update statistics
        encounter_sets = list(self.project.encounter_sets)
        self.encounter_count_label.setText(str(len(encounter_sets)))
        self.card_count_label.setText(str(len(self.project.get_all_cards())))
        self.player_card_count_label.setText(str(len(self.project.player_cards)))

        self._updating = False

    def on_field_changed(self, field_name):
        """Handle field change"""
        if self._updating:
            return

        if field_name == 'name':
            self.project.data['name'] = self.name_input.text()
        elif field_name == 'code':
            self.project.data['code'] = self.code_input.text()
        elif field_name == 'default_copyright':
            self.project.data['default_copyright'] = self.copyright_input.text()
        elif field_name == 'icon':
            self.project.data['icon'] = self.icon_input.text()

        self.project.dirty = True
        self.data_changed.emit()

    def browse_icon(self):
        """Browse for icon file"""
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Select Icon",
            str(Path.home()),
            "Images (*.png *.jpg *.jpeg *.webp)"
        )
        if file_path:
            self.icon_input.setText(file_path)

    def add_encounter_set(self):
        """Add a new encounter set to the project"""
        name, ok = QInputDialog.getText(
            self,
            "New Encounter Set",
            "Enter encounter set name:"
        )
        if ok and name:
            try:
                self.project.add_encounter_set(name)
                self.load_data()  # Refresh stats
                QMessageBox.information(
                    self,
                    "Success",
                    f"Encounter set '{name}' created."
                )
            except Exception as e:
                QMessageBox.critical(
                    self,
                    "Error",
                    f"Failed to create encounter set: {e}"
                )

    def load_thumbnails(self):
        """Load card thumbnails in background"""
        if not self.card_renderer:
            return

        cards = self.project.get_all_cards()

        for index, card in enumerate(cards):
            try:
                # Render thumbnail
                front_image, _ = self.card_renderer.get_card_textures(card, bleed=False)

                # Convert to QPixmap
                front_image.seek(0)
                image_data = front_image.read()
                qimage = QImage.fromData(image_data)
                pixmap = QPixmap.fromImage(qimage)

                # Scale to thumbnail size
                thumbnail = pixmap.scaled(
                    150, 210,
                    Qt.KeepAspectRatio,
                    Qt.SmoothTransformation
                )

                # Emit signal to add thumbnail in main thread
                self.thumbnail_ready.emit(thumbnail, card.name, index)

            except Exception as e:
                print(f"Error rendering thumbnail for {card.name}: {e}")

    @Slot(QPixmap, str, int)
    def _add_thumbnail(self, pixmap, card_name, index):
        """Add thumbnail to grid (called from main thread via signal)"""
        cols = 4
        row = index // cols
        col = index % cols

        label = QLabel()
        label.setPixmap(pixmap)
        label.setToolTip(card_name)
        label.setAlignment(Qt.AlignCenter)

        self.thumbnail_grid.addWidget(label, row, col)



"""
Strange Eons Converter Dialog
"""
import multiprocessing
from pathlib import Path

from PySide6.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QFormLayout,
    QLineEdit, QPushButton, QLabel, QFileDialog,
    QDialogButtonBox, QMessageBox,
    QPlainTextEdit
)
from PySide6.QtCore import Qt, QTimer


def _run_conversion_subprocess(java_path, jar_path, project_path, output_path, log_queue):
    """
    Target function for multiprocessing.Process.

    Runs the Strange Eons conversion in an isolated subprocess.
    When JPype/Strange Eons terminates, only this subprocess dies.
    """
    import sys

    # Redirect stdout/stderr to the queue
    class QueueWriter:
        def __init__(self, queue):
            self.queue = queue

        def write(self, text):
            if text.strip():
                self.queue.put(text)

        def flush(self):
            pass

    sys.stdout = QueueWriter(log_queue)
    sys.stderr = QueueWriter(log_queue)

    try:
        from shoggoth.strange_eons import run_conversion
        run_conversion(java_path, jar_path, project_path, output_path)
        log_queue.put('__SUCCESS__')
    except Exception as e:
        log_queue.put(f'Error: {e}')
        log_queue.put('__FAILED__')
    finally:
        log_queue.put('__DONE__')


class StrangeEonsConverterDialog(QDialog):
    """Dialog for converting Strange Eons projects"""

    def __init__(self, settings, parent=None):
        super().__init__(parent)
        self.settings = settings
        self.setWindowTitle("Convert Strange Eons Project")
        self.setMinimumWidth(500)

        layout = QVBoxLayout(self)

        # Form layout for inputs
        form = QFormLayout()

        # Java path
        java_layout = QHBoxLayout()
        self.java_input = QLineEdit()
        self.java_input.setText(self.settings.get('java_path', 'java'))
        self.java_input.setPlaceholderText("java (uses system default if empty)")
        java_browse = QPushButton("Browse...")
        java_browse.clicked.connect(self._browse_java)
        java_layout.addWidget(self.java_input)
        java_layout.addWidget(java_browse)
        form.addRow("Java Path:", java_layout)

        # Strange Eons JAR path
        jar_layout = QHBoxLayout()
        self.jar_input = QLineEdit()
        self.jar_input.setText(self.settings.get('strange_eons_path', ''))
        self.jar_input.setPlaceholderText("Path to strange-eons.jar")
        jar_browse = QPushButton("Browse...")
        jar_browse.clicked.connect(self._browse_jar)
        jar_layout.addWidget(self.jar_input)
        jar_layout.addWidget(jar_browse)
        form.addRow("Strange Eons JAR:", jar_layout)

        # SE Project path
        project_layout = QHBoxLayout()
        self.project_input = QLineEdit()
        self.project_input.setPlaceholderText("Path to Strange Eons project folder")
        project_browse = QPushButton("Browse...")
        project_browse.clicked.connect(self._browse_project)
        project_layout.addWidget(self.project_input)
        project_layout.addWidget(project_browse)
        form.addRow("SE Project:", project_layout)

        # Output folder
        output_layout = QHBoxLayout()
        self.output_input = QLineEdit()
        self.output_input.setPlaceholderText("Output folder for converted project")
        output_browse = QPushButton("Browse...")
        output_browse.clicked.connect(self._browse_output)
        output_layout.addWidget(self.output_input)
        output_layout.addWidget(output_browse)
        form.addRow("Output Folder:", output_layout)

        layout.addLayout(form)

        # Info label
        info = QLabel(
            "<i>Note: The conversion process will run in a separate process. "
            "This may take a while depending on the project size.</i>"
        )
        info.setWordWrap(True)
        layout.addWidget(info)

        # Buttons
        buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        buttons.accepted.connect(self._start_conversion)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

    def _browse_java(self):
        path, _ = QFileDialog.getOpenFileName(
            self, "Select Java Executable",
            filter="Executables (*);;All Files (*)"
        )
        if path:
            self.java_input.setText(path)

    def _browse_jar(self):
        path, _ = QFileDialog.getOpenFileName(
            self, "Select Strange Eons JAR",
            filter="JAR Files (*.jar);;All Files (*)"
        )
        if path:
            self.jar_input.setText(path)

    def _browse_project(self):
        path = QFileDialog.getExistingDirectory(
            self, "Select Strange Eons Project Folder"
        )
        if path:
            self.project_input.setText(path)
            # Auto-fill output if empty
            if not self.output_input.text():
                project_path = Path(path)
                self.output_input.setText(str(project_path.parent / f"{project_path.name}_converted"))

    def _browse_output(self):
        path = QFileDialog.getExistingDirectory(
            self, "Select Output Folder"
        )
        if path:
            self.output_input.setText(path)

    def _validate(self):
        """Validate inputs"""
        if not self.jar_input.text():
            QMessageBox.warning(self, "Error", "Please specify the Strange Eons JAR path.")
            return False

        jar_path = Path(self.jar_input.text())
        if not jar_path.exists():
            QMessageBox.warning(self, "Error", f"Strange Eons JAR not found: {jar_path}")
            return False

        if not self.project_input.text():
            QMessageBox.warning(self, "Error", "Please specify the Strange Eons project folder.")
            return False

        project_path = Path(self.project_input.text())
        if not project_path.exists():
            QMessageBox.warning(self, "Error", f"Project folder not found: {project_path}")
            return False

        if not self.output_input.text():
            QMessageBox.warning(self, "Error", "Please specify the output folder.")
            return False

        return True

    def _start_conversion(self):
        """Start the conversion process"""
        if not self._validate():
            return

        # Save settings for future use
        self.settings['java_path'] = self.java_input.text()
        self.settings['strange_eons_path'] = self.jar_input.text()

        # Store paths for the caller
        self.java_path = self.java_input.text() or 'java'
        self.jar_path = self.jar_input.text()
        self.project_path = self.project_input.text()
        self.output_path = self.output_input.text()

        self.accept()


class ConversionProgressDialog(QDialog):
    """Progress dialog for the conversion process"""

    def __init__(self, java_path, jar_path, project_path, output_path, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Converting Project...")
        self.setMinimumSize(600, 400)
        self.setModal(True)

        # Don't allow closing while running
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowCloseButtonHint)

        self.output_path = output_path
        self.process = None
        self.log_queue = None
        self.poll_timer = None
        self.success = False

        layout = QVBoxLayout(self)

        self.status_label = QLabel("Starting conversion...")
        layout.addWidget(self.status_label)

        # Log output area
        self.log_view = QPlainTextEdit()
        self.log_view.setReadOnly(True)
        self.log_view.setLineWrapMode(QPlainTextEdit.NoWrap)
        layout.addWidget(self.log_view, 1)  # stretch factor 1

        # Cancel button
        self.cancel_btn = QPushButton("Cancel")
        self.cancel_btn.clicked.connect(self._cancel)
        layout.addWidget(self.cancel_btn)

        # Start the process
        QTimer.singleShot(100, lambda: self._run_conversion(
            java_path, jar_path, project_path, output_path
        ))

    def _run_conversion(self, java_path, jar_path, project_path, output_path):
        """Run the conversion in a subprocess using multiprocessing"""
        # Log configuration
        self._append_log("=== Strange Eons Conversion ===")
        self._append_log(f"Java: {java_path}")
        self._append_log(f"JAR: {jar_path}")
        self._append_log(f"Project: {project_path}")
        self._append_log(f"Output: {output_path}")
        self._append_log("")

        self.status_label.setText("Converting... (this may take a while)")

        # Use multiprocessing to run conversion in isolated subprocess
        # This works with frozen executables and isolates JPype termination
        ctx = multiprocessing.get_context('spawn')
        self.log_queue = ctx.Queue()

        self.process = ctx.Process(
            target=_run_conversion_subprocess,
            args=(java_path, jar_path, project_path, output_path, self.log_queue)
        )
        self.process.start()

        # Poll the queue for log output
        self.poll_timer = QTimer(self)
        self.poll_timer.timeout.connect(self._poll_queue)
        self.poll_timer.start(100)

    def _append_log(self, text):
        """Append text to log view and scroll to bottom"""
        self.log_view.appendPlainText(text.rstrip('\n'))
        # Scroll to bottom
        scrollbar = self.log_view.verticalScrollBar()
        scrollbar.setValue(scrollbar.maximum())

    def _poll_queue(self):
        """Poll the multiprocessing queue for log messages"""
        try:
            while not self.log_queue.empty():
                msg = self.log_queue.get_nowait()
                if msg == '__DONE__':
                    self._on_finished()
                    return
                elif msg == '__SUCCESS__':
                    self.success = True
                elif msg == '__FAILED__':
                    self.success = False
                else:
                    self._append_log(msg)
        except Exception:
            pass

        # Check if process died unexpectedly
        if self.process and not self.process.is_alive():
            self._on_finished()

    def _on_finished(self):
        """Handle process completion"""
        if self.poll_timer:
            self.poll_timer.stop()
            self.poll_timer = None

        exit_code = self.process.exitcode if self.process else -1
        self.process = None

        self._append_log(f"\n=== Process finished with exit code {exit_code} ===")

        # Check if output was created
        output_project = Path(self.output_path) / "cards.json"
        if output_project.exists():
            self._append_log(f"Success: {output_project}")
            self.success = True
            self.accept()
        else:
            self.status_label.setText("Conversion completed but no project was created.")
            self._append_log(f"Expected output: {output_project}")
            self.cancel_btn.setText("Close")
            # Re-enable close button
            self.setWindowFlags(self.windowFlags() | Qt.WindowCloseButtonHint)
            self.show()

    def _cancel(self):
        if self.poll_timer:
            self.poll_timer.stop()
        if self.process and self.process.is_alive():
            self.process.terminate()
            self.process.join(timeout=1)
            if self.process.is_alive():
                self.process.kill()
        self.reject()

    def closeEvent(self, event):
        if self.process and self.process.is_alive():
            event.ignore()
        else:
            super().closeEvent(event)


def run_strange_eons_conversion(settings, parent=None):
    """
    Run the complete Strange Eons conversion workflow.

    Returns the path to the converted project if successful, None otherwise.
    """
    # Step 1: Show converter dialog
    dialog = StrangeEonsConverterDialog(settings, parent)
    if dialog.exec() != QDialog.Accepted:
        return None

    # Step 2: Run conversion with progress
    progress = ConversionProgressDialog(
        dialog.java_path,
        dialog.jar_path,
        dialog.project_path,
        dialog.output_path,
        parent
    )

    if progress.exec() != QDialog.Accepted or not progress.success:
        return None

    # Return path to open directly
    output_project = Path(dialog.output_path) / "cards.json"
    return str(output_project)



"""
Main application entry point for Shoggoth with PySide6
"""
import sys
from PySide6.QtWidgets import QApplication
from PySide6.QtCore import Qt

from shoggoth.main_window import ShoggothMainWindow


def main():
    """Main application entry point"""
    # Enable high DPI scaling
    QApplication.setHighDpiScaleFactorRoundingPolicy(
        Qt.HighDpiScaleFactorRoundingPolicy.PassThrough
    )
    
    app = QApplication(sys.argv)
    app.setApplicationName("Shoggoth")
    app.setOrganizationName("Shoggoth")
    
    # Set application style
    app.setStyle("Fusion")
    
    # Create and show main window
    window = ShoggothMainWindow()
    window.show()
    
    sys.exit(app.exec())


if __name__ == "__main__":
    main()


"""
Auto-update functionality for Shoggoth

Supports both PyPI and GitHub releases based on installation type.
"""
import sys
import json
import logging
import threading
import requests
import tempfile
import subprocess
import platform
from pathlib import Path
from enum import Enum
from dataclasses import dataclass
from typing import Optional

from PySide6.QtCore import QObject, Signal, QProcess, QTimer
from PySide6.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QFormLayout,
    QLabel, QPushButton, QTextBrowser, QProgressBar,
    QPlainTextEdit, QMessageBox, QApplication
)

# Constants
GITHUB_API_URL = "https://api.github.com/repos/tokeeto/shoggoth/releases/latest"
PYPI_API_URL = "https://pypi.org/pypi/shoggoth/json"

logger = logging.getLogger(__name__)


class InstallationType(Enum):
    """How Shoggoth was installed"""
    PYPI = "pypi"           # pip install shoggoth
    BINARY = "binary"       # PyInstaller frozen exe
    DEVELOPMENT = "dev"     # Running from source (uv run, pip -e)


@dataclass
class VersionInfo:
    """Information about an available version"""
    version: str
    download_url: Optional[str] = None
    release_notes: Optional[str] = None
    published_at: Optional[str] = None


def get_current_version() -> str:
    """Get current version from package metadata"""
    try:
        from importlib.metadata import version
        return version("shoggoth")
    except Exception:
        return "unknown"


def detect_installation_type() -> InstallationType:
    """Detect how Shoggoth was installed"""
    # Check for PyInstaller frozen executable
    if getattr(sys, 'frozen', False):
        return InstallationType.BINARY

    # Check if running from a pip-installed package
    try:
        from importlib.metadata import distribution
        dist = distribution('shoggoth')

        # Check if it's an editable install (development)
        try:
            direct_url_text = dist.read_text('direct_url.json')
            if direct_url_text:
                direct_url = json.loads(direct_url_text)
                if direct_url.get('dir_info', {}).get('editable', False):
                    return InstallationType.DEVELOPMENT
        except (FileNotFoundError, json.JSONDecodeError):
            pass

        return InstallationType.PYPI
    except Exception as e:
        # Package not found via importlib.metadata - likely development
        return InstallationType.DEVELOPMENT


def compare_versions(current: str, latest: str) -> bool:
    """
    Compare semver versions, return True if latest > current.

    Handles version strings like "0.1.5", "v0.2.0", "1.0.0-beta".
    """
    def normalize(v: str) -> tuple:
        # Remove leading 'v' if present
        v = v.lstrip('v')
        # Split on common separators and convert to comparable tuple
        parts = v.replace('-', '.').replace('_', '.').split('.')
        result = []
        for part in parts:
            try:
                result.append((0, int(part)))  # Numeric parts sort first
            except ValueError:
                result.append((1, part))  # String parts sort after
        return tuple(result)

    try:
        # Try using packaging library if available
        from packaging.version import Version
        return Version(latest.lstrip('v')) > Version(current.lstrip('v'))
    except ImportError:
        # Fallback to simple comparison
        return normalize(latest) > normalize(current)
    except Exception:
        # If all else fails, simple string comparison
        return latest.lstrip('v') != current.lstrip('v')


class UpdateChecker(QObject):
    """Background update checker with thread-safe signals"""

    # Signals for thread-safe communication
    update_available = Signal(object)  # VersionInfo
    check_complete = Signal(bool)      # has_update
    check_failed = Signal(str)         # error message

    def __init__(self, parent=None):
        super().__init__(parent)
        self.current_version = get_current_version()
        self.installation_type = detect_installation_type()

    def check_for_updates(self):
        """Start background update check (spawns thread)"""
        thread = threading.Thread(target=self._do_check, daemon=True)
        thread.start()

    def _do_check(self):
        """Perform the actual update check (runs in background thread)"""
        try:
            if self.installation_type == InstallationType.BINARY:
                version_info = self._check_github_releases()
            else:
                version_info = self._check_pypi()

            if version_info and compare_versions(self.current_version, version_info.version):
                self.update_available.emit(version_info)
                self.check_complete.emit(True)
            else:
                self.check_complete.emit(False)
        except Exception as e:
            logger.warning(f"Update check failed: {e}")
            self.check_failed.emit(str(e))

    def _check_pypi(self) -> Optional[VersionInfo]:
        """Query PyPI JSON API for latest version"""
        try:
            response = requests.get(
                PYPI_API_URL,
                headers={'Accept': 'application/json', 'User-Agent': 'Shoggoth-Updater'}
            )
            data = response.json()

            info = data.get('info', {})
            return VersionInfo(
                version=info.get('version', ''),
                release_notes=info.get('summary', ''),
            )
        except Exception as e:
            logger.warning(f"PyPI check failed: {e}")
            raise

    def _check_github_releases(self) -> Optional[VersionInfo]:
        """Query GitHub Releases API for latest version"""
        try:
            response = requests.get(
                GITHUB_API_URL,
                headers={
                    'Accept': 'application/vnd.github.v3+json',
                    'User-Agent': 'Shoggoth-Updater'
                }
            )
            data = response.json()

            # Find appropriate asset for this platform
            download_url = None
            assets = data.get('assets', [])
            system = platform.system().lower()

            for asset in assets:
                name = asset.get('name', '').lower()
                # Match platform-specific installers
                if system == 'windows' and 'win' in name:
                    download_url = asset.get('browser_download_url')
                    break
                elif system == 'darwin' and 'mac' in name:
                    download_url = asset.get('browser_download_url')
                    break

            return VersionInfo(
                version=data.get('tag_name', '').lstrip('v'),
                download_url=download_url,
                release_notes=data.get('body', ''),
                published_at=data.get('published_at', ''),
            )
        except Exception as e:
            logger.warning(f"GitHub check failed: {e}")
            raise


class UpdateDialog(QDialog):
    """Dialog shown when an update is available"""

    # Result codes
    UPDATE_NOW = 1
    NOT_NOW = 2
    SKIP_VERSION = 3

    def __init__(self, version_info: VersionInfo, current_version: str, parent=None):
        super().__init__(parent)
        self.version_info = version_info
        self.current_version = current_version
        self.result_action = self.NOT_NOW

        self.setWindowTitle("Update Available")
        self.setMinimumWidth(500)
        self.setMinimumHeight(300)

        self.setup_ui()

    def setup_ui(self):
        """Build the dialog layout"""
        layout = QVBoxLayout()

        # Header
        header = QLabel("A new version of Shoggoth is available!")
        header.setStyleSheet("font-size: 14pt; font-weight: bold; margin-bottom: 10px;")
        layout.addWidget(header)

        # Version info
        version_layout = QFormLayout()
        version_layout.addRow("Current version:", QLabel(self.current_version))
        version_layout.addRow("New version:", QLabel(f"<b>{self.version_info.version}</b>"))
        layout.addLayout(version_layout)

        # Release notes
        if self.version_info.release_notes:
            notes_label = QLabel("Release Notes:")
            notes_label.setStyleSheet("margin-top: 10px; font-weight: bold;")
            layout.addWidget(notes_label)

            notes_browser = QTextBrowser()
            notes_browser.setMarkdown(self.version_info.release_notes)
            notes_browser.setMaximumHeight(200)
            layout.addWidget(notes_browser)

        layout.addStretch()

        # Buttons
        button_layout = QHBoxLayout()

        update_btn = QPushButton("Update Now")
        update_btn.setDefault(True)
        update_btn.clicked.connect(self._on_update_now)
        button_layout.addWidget(update_btn)

        later_btn = QPushButton("Not Now")
        later_btn.clicked.connect(self._on_not_now)
        button_layout.addWidget(later_btn)

        skip_btn = QPushButton("Skip This Version")
        skip_btn.clicked.connect(self._on_skip_version)
        button_layout.addWidget(skip_btn)

        layout.addLayout(button_layout)

        self.setLayout(layout)

    def _on_update_now(self):
        self.result_action = self.UPDATE_NOW
        self.accept()

    def _on_not_now(self):
        self.result_action = self.NOT_NOW
        self.reject()

    def _on_skip_version(self):
        self.result_action = self.SKIP_VERSION
        self.reject()


class UpdateProgressDialog(QDialog):
    """Shows update download/install progress"""

    def __init__(self, version_info: VersionInfo, installation_type: InstallationType, parent=None):
        super().__init__(parent)
        self.version_info = version_info
        self.installation_type = installation_type
        self.process = None
        self.download_path = None

        self.setWindowTitle("Updating Shoggoth")
        self.setMinimumWidth(500)
        self.setMinimumHeight(300)
        self.setModal(True)

        self.setup_ui()

    def setup_ui(self):
        """Build the dialog layout"""
        layout = QVBoxLayout()

        # Status label
        self.status_label = QLabel("Preparing update...")
        self.status_label.setStyleSheet("font-weight: bold;")
        layout.addWidget(self.status_label)

        # Progress bar
        self.progress_bar = QProgressBar()
        self.progress_bar.setRange(0, 100)
        self.progress_bar.setValue(0)
        layout.addWidget(self.progress_bar)

        # Output log
        self.output_log = QPlainTextEdit()
        self.output_log.setReadOnly(True)
        self.output_log.setMaximumHeight(150)
        layout.addWidget(self.output_log)

        layout.addStretch()

        # Buttons
        button_layout = QHBoxLayout()
        button_layout.addStretch()

        self.cancel_btn = QPushButton("Cancel")
        self.cancel_btn.clicked.connect(self._on_cancel)
        button_layout.addWidget(self.cancel_btn)

        self.close_btn = QPushButton("Close")
        self.close_btn.clicked.connect(self.accept)
        self.close_btn.setVisible(False)
        button_layout.addWidget(self.close_btn)

        layout.addLayout(button_layout)

        self.setLayout(layout)

    def start_update(self):
        """Begin the update process"""
        if self.installation_type == InstallationType.PYPI:
            self._start_pip_upgrade()
        elif self.installation_type == InstallationType.BINARY:
            self._start_binary_download()

    def _start_pip_upgrade(self):
        """Run pip upgrade for PyPI installation"""
        self.status_label.setText("Installing update via pip...")
        self.progress_bar.setRange(0, 0)  # Indeterminate progress

        self.process = QProcess(self)
        self.process.readyReadStandardOutput.connect(self._on_stdout)
        self.process.readyReadStandardError.connect(self._on_stderr)
        self.process.finished.connect(self._on_pip_finished)

        # Use sys.executable to ensure correct Python
        self.process.start(sys.executable, ["-m", "pip", "install", "--upgrade", 'shoggoth'])

    def _on_stdout(self):
        """Handle stdout from pip"""
        if self.process:
            data = self.process.readAllStandardOutput().data().decode('utf-8', errors='replace')
            self.output_log.appendPlainText(data.strip())

    def _on_stderr(self):
        """Handle stderr from pip"""
        if self.process:
            data = self.process.readAllStandardError().data().decode('utf-8', errors='replace')
            self.output_log.appendPlainText(data.strip())

    def _on_pip_finished(self, exit_code, exit_status):
        """Handle pip completion"""
        self.progress_bar.setRange(0, 100)
        self.progress_bar.setValue(100)
        self.cancel_btn.setVisible(False)
        self.close_btn.setVisible(True)

        if exit_code == 0:
            self.status_label.setText("Update complete! Please restart Shoggoth.")
            self.output_log.appendPlainText("\n--- Update successful ---")
        else:
            self.status_label.setText("Update failed. See log for details.")
            self.output_log.appendPlainText(f"\n--- Update failed (exit code {exit_code}) ---")
            self.output_log.appendPlainText("You can try manually: pip install --upgrade shoggoth")

    def _start_binary_download(self):
        """Download binary update"""
        if not self.version_info.download_url:
            self.status_label.setText("No download available for your platform.")
            self.output_log.appendPlainText("Could not find a compatible download.")
            self.output_log.appendPlainText(f"Please visit: https://github.com/tokeeto/shoggoth/releases/latest")
            self.cancel_btn.setVisible(False)
            self.close_btn.setVisible(True)
            return

        self.status_label.setText("Downloading update...")

        # Download in background thread
        thread = threading.Thread(target=self._download_binary, daemon=True)
        thread.start()

    def _download_binary(self):
        """Download binary file (runs in background thread)"""
        try:
            # Determine filename from URL
            url = self.version_info.download_url
            if not url:
                raise
            filename = url.split('/')[-1]

            # Download to temp directory
            temp_dir = Path(tempfile.gettempdir())
            self.download_path = temp_dir / filename

            def report_progress(block_num, block_size, total_size):
                if total_size > 0:
                    downloaded = block_num * block_size
                    percent = min(100, int(downloaded * 100 / total_size))
                    # Update UI from main thread
                    QTimer.singleShot(0, lambda: self._update_download_progress(percent, downloaded, total_size))

            import urllib.request
            urllib.request.urlretrieve(url, self.download_path, reporthook=report_progress)

            # Download complete - update UI from main thread
            QTimer.singleShot(0, self._on_download_complete)

        except Exception as e:
            QTimer.singleShot(0, lambda: self._on_download_error(str(e)))

    def _update_download_progress(self, percent, downloaded, total):
        """Update download progress (called on main thread)"""
        self.progress_bar.setValue(percent)
        mb_downloaded = downloaded / (1024 * 1024)
        mb_total = total / (1024 * 1024)
        self.status_label.setText(f"Downloading... {mb_downloaded:.1f} / {mb_total:.1f} MB")

    def _on_download_complete(self):
        """Handle download completion"""
        self.progress_bar.setValue(100)
        self.status_label.setText("Download complete!")
        self.cancel_btn.setVisible(False)
        self.close_btn.setVisible(True)

        self.output_log.appendPlainText(f"Downloaded to: {self.download_path}")
        self.output_log.appendPlainText("\nClick 'Run Installer' to install the update.")

        # Add run installer button
        run_btn = QPushButton("Run Installer")
        run_btn.clicked.connect(self._run_installer)
        self.layout().itemAt(self.layout().count() - 1).layout().insertWidget(0, run_btn)

    def _on_download_error(self, error: str):
        """Handle download error"""
        self.status_label.setText("Download failed")
        self.output_log.appendPlainText(f"Error: {error}")
        self.output_log.appendPlainText(f"\nPlease download manually from:")
        self.output_log.appendPlainText(f"https://github.com/tokeeto/shoggoth/releases/latest")
        self.cancel_btn.setVisible(False)
        self.close_btn.setVisible(True)

    def _run_installer(self):
        """Run the downloaded installer"""
        if not self.download_path or not self.download_path.exists():
            return

        try:
            if sys.platform == 'win32':
                import os
                os.startfile(str(self.download_path))
            elif sys.platform == 'darwin':
                subprocess.Popen(['open', str(self.download_path)])
            else:
                # Linux - try to make executable and run
                self.download_path.chmod(self.download_path.stat().st_mode | 0o111)
                subprocess.Popen([str(self.download_path)])

            # Quit the application
            QApplication.quit()
        except Exception as e:
            QMessageBox.critical(
                self,
                "Error",
                f"Could not run installer:\n{e}\n\nPlease run it manually from:\n{self.download_path}"
            )

    def _on_cancel(self):
        """Cancel the update"""
        if self.process and self.process.state() != QProcess.NotRunning:
            self.process.kill()
        self.reject()


class UpdateManager(QObject):
    """Coordinates update checking and UI"""

    def __init__(self, settings_manager, parent=None):
        super().__init__(parent)
        self.settings = settings_manager
        self.parent_widget = parent
        self.checker = UpdateChecker(self)

        # Connect signals
        self.checker.update_available.connect(self._on_update_available)
        self.checker.check_complete.connect(self._on_check_complete)
        self.checker.check_failed.connect(self._on_check_failed)

        self._is_manual_check = False

    def should_check_for_updates(self) -> bool:
        """Check if we should perform an automatic update check"""
        # Skip in development mode
        if self.checker.installation_type == InstallationType.DEVELOPMENT:
            logger.info("Skipping update check in development mode")
            return False

        # Check if auto-updates are enabled
        if not self.settings.getboolean('Shoggoth', 'auto_check_updates', True):
            return False

        return True

    def check_for_updates_async(self):
        """Trigger background update check (for startup)"""
        self._is_manual_check = False
        self.checker.check_for_updates()

    def check_for_updates_manual(self):
        """Trigger manual update check (from menu)"""
        self._is_manual_check = True

        # For development mode, show a message
        if self.checker.installation_type == InstallationType.DEVELOPMENT:
            QMessageBox.information(
                self.parent_widget,
                "Development Mode",
                "Update checking is disabled in development mode.\n\n"
                "You are running Shoggoth from source code."
            )
            return

        self.checker.check_for_updates()

    def _on_check_complete(self, has_update: bool):
        """Handle check complete signal"""
        if not has_update and self._is_manual_check:
            QMessageBox.information(
                self.parent_widget,
                "Up to Date",
                f"Shoggoth is already up to date.\n\nCurrent version: {self.checker.current_version}"
            )

    def _on_update_available(self, version_info: VersionInfo):
        """Handle update available signal"""
        # Check if this version should be skipped
        skipped = self.settings.get('Shoggoth', 'skipped_version', '')
        if not self._is_manual_check and skipped == version_info.version:
            logger.info(f"Skipping update notification for version {version_info.version}")
            return

        # Show update dialog
        dialog = UpdateDialog(version_info, self.checker.current_version, self.parent_widget)
        dialog.exec()

        if dialog.result_action == UpdateDialog.UPDATE_NOW:
            self._perform_update(version_info)
        elif dialog.result_action == UpdateDialog.SKIP_VERSION:
            self.settings.set('Shoggoth', 'skipped_version', version_info.version)
            self.settings.save()

    def _on_check_failed(self, error: str):
        """Handle check failed signal"""
        if self._is_manual_check:
            QMessageBox.warning(
                self.parent_widget,
                "Update Check Failed",
                f"Could not check for updates:\n{error}"
            )

    def _perform_update(self, version_info: VersionInfo):
        """Execute the update"""
        dialog = UpdateProgressDialog(version_info, self.checker.installation_type, self.parent_widget)
        dialog.show()
        dialog.start_update()
        dialog.exec()



import json
import os
import shoggoth
from shoggoth import files
from copy import deepcopy
from pathlib import Path

wrapper_template = {
    "SaveName": "",
    "Date": "",
    "VersionNumber": "",
    "GameMode": "",
    "GameType": "",
    "GameComplexity": "",
    "Tags": [],
    "Gravity": 0.5,
    "PlayArea": 0.5,
    "Table": "",
    "Sky": "",
    "Note": "",
    "TabStates": {},
    "LuaScript": "",
    "LuaScriptState": "",
    "XmlUI": "",
    "ObjectStates": []
}

card_template = {
    "CardID": 552100,
    "CustomDeck": {
        "5521": {
            "BackIsHidden": True,
            "BackURL": "https://steamusercontent-a.akamaihd.net/ugc/2342503777940352139/A2D42E7E5C43D045D72CE5CFC907E4F886C8C690/",
            "FaceURL": "",
            "NumHeight": 1,
            "NumWidth": 1,
            "Type": 0,
            "UniqueBack": True
        }
    },
    "Description": "Card 1",
    "GMNotes_path": None,
    "GUID": "427b4e28",
    "Name": "Card",
    "Nickname": "A card",
    "Tags": ["Asset", "PlayerCard"],
    "Transform": {
        "posX": 0,
        "posY": 0,
        "posZ": 0,
        "rotX": 0,
        "rotY": 270,
        "rotZ": 0,
        "scaleX": 1,
        "scaleY": 1,
        "scaleZ": 1
    }
}

campaign_box_template = {
  "AltLookAngle": {
    "x": 0,
    "y": 0,
    "z": 0
  },
  "Autoraise": True,
  "ColorDiffuse": {
    "a": 0.27451,
    "b": 1,
    "g": 1,
    "r": 1
  },
  "CustomMesh": {
    "CastShadows": True,
    "ColliderURL": "",
    "Convex": True,
    "CustomShader": {
      "FresnelStrength": 0,
      "SpecularColor": {
        "b": 1,
        "g": 1,
        "r": 1
      },
      "SpecularIntensity": 0,
      "SpecularSharpness": 2
    },
    "DiffuseURL": "https://steamusercontent-a.akamaihd.net/ugc/2038486699957628515/8202EA3F06FDDD807A34BD6F62FE2E0A0723B8CD/",
    "MaterialIndex": 3,
    "MeshURL": "https://steamusercontent-a.akamaihd.net/ugc/62583916778515295/AFB8F257CE1E4973F4C06160A2E156C147AEE1E3/",
    "NormalURL": "",
    "TypeIndex": 0
  },
  "Description": "",
  "DragSelectable": True,
  "GMNotes": "{\n  \"filename\": \"the_scarlet_keys\"}",
  "GUID": "300fcc",
  "Grid": True,
  "GridProjection": False,
  "Hands": False,
  "HideWhenFaceDown": False,
  "IgnoreFoW": False,
  "LayoutGroupSortIndex": 0,
  "Locked": False,
  "LuaScript": "require(\"core/DownloadBox\")",
  "LuaScriptState": "",
  "MeasureMovement": False,
  "Name": "Custom_Model",
  "Nickname": "The Scarlet Keys",
  "Snap": True,
  "Sticky": True,
  "Tags": [
    "CampaignBox"
  ],
  "Tooltip": True,
  "Transform": {
    "posX": 60,
    "posY": 1.481,
    "posZ": -63.33,
    "rotX": 0,
    "rotY": 270,
    "rotZ": 0,
    "scaleX": 1,
    "scaleY": 0.14,
    "scaleZ": 1
  },
  "Value": 0,
  "XmlUI": ""
}

def get_image_path(card):
    export = Path(shoggoth.app.current_project.file_path).parent / f'export_of_{shoggoth.app.current_project.name}' / f'{card.name}_front.png'
    return export


def card_to_tts(card):
    data = deepcopy(card_template)
    data['CustomDeck']['5521']['FaceURL'] = f'file:///{get_image_path(card)}'
    data['Description'] = card.name
    data['GUID'] = card.id
    data['Tags'] = ["Asset", "PlayerCard"]
    return data


def export_card(card):
    wrapper = deepcopy(wrapper_template)
    data = card_to_tts(card)
    wrapper['ObjectStates'].append(data)

    if files.tts_dir:
        output_path = files.tts_dir / f"{card.name}.json"
    else:
        output_path = Path(shoggoth.app.current_project.file_path).parent / f"{card.name}.json"

    with open(output_path, 'w') as file:
        json.dump(wrapper, file, indent=4)


def export_campaign(expansion):
    wrapper = deepcopy(wrapper_template)
    for encounter in expansion.encounter_sets:
        encounter_wrapper = deepcopy(wrapper_template)
        for card in encounter.cards:
            encounter_wrapper['ObjectStates'].append(card_to_tts(card))
        wrapper['ObjectStates'].append(encounter_wrapper)

    if files.tts_dir:
        output_path = files.tts_dir / f"{shoggoth.app.current_project.name} campaign.json"
    else:
        output_path = Path(shoggoth.app.current_project.file_path).parent / f"{shoggoth.app.current_project.name} campaign.json"

    with open(output_path, 'w') as file:
        json.dump(wrapper, file, indent=4)


def export_player_cards(cards):
    wrapper = deepcopy(wrapper_template)
    for card in cards:
        wrapper['ObjectStates'].append(card_to_tts(card))

    if files.tts_dir:
        output_path = files.tts_dir / f"{shoggoth.app.current_project.name} player cards.json"
    else:
        output_path = Path(shoggoth.app.current_project.file_path).parent / f"{shoggoth.app.current_project.name} player cards.json"

    with open(output_path, 'w') as file:
        json.dump(wrapper, file, indent=4)



"""
Face editors for different card types - Consolidated
"""
from PySide6.QtWidgets import (
    QBoxLayout, QSpacerItem, QWidget, QVBoxLayout, QHBoxLayout, QFormLayout,
    QLineEdit, QTextEdit, QPlainTextEdit, QPushButton, QComboBox,
    QLabel, QFileDialog, QGridLayout, QGroupBox, QCompleter, QSpinBox
)
from PySide6.QtCore import Signal, Qt, QSize
from pathlib import Path
import json

# Import helper widgets from editors.py
from shoggoth import editors


# All available card types
ALL_CARD_TYPES = [
    'asset', 'event', 'skill',
    'investigator', 'investigator_back',
    'enemy', 'treachery', 'location', 'location_back',
    'act', 'act_back', 'agenda', 'agenda_back',
    'scenario', 'chaos',
    'customizable', 'story',
    'player', 'encounter',
]
FULLART_CARD_TYPES = [
    'fullart_asset',
    'fullart_event',
    'fullart_skill',
    'fullart_investigator',
    'fullart_enemy',
    'fullart_treachery',
]


class NoScrollComboBox(QComboBox):
    """ComboBox that ignores wheel events when not focused."""

    def wheelEvent(self, event):
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()


class SlotComboBox(NoScrollComboBox):
    """ComboBox that displays available slot types with icons."""

    def __init__(self, parent=None):
        super().__init__(parent)
        from shoggoth.files import overlay_dir
        from PySide6.QtGui import QIcon

        self.setIconSize(QSize(24, 24))

        # Add empty option first
        self.addItem('—', userData=None)

        # Discover slot files dynamically
        if overlay_dir.exists():
            slot_files = sorted(overlay_dir.glob('slot_*.png'))
            for slot_file in slot_files:
                # Extract name from "slot_<name>.png"
                name = slot_file.stem[5:]  # Remove "slot_" prefix
                icon = QIcon(str(slot_file))
                self.addItem(icon, name, userData=name)

    def setCurrentSlot(self, slot_name):
        """Set current selection by slot name"""
        if not slot_name:
            self.setCurrentIndex(0)
        else:
            for i in range(self.count()):
                if self.itemData(i) == slot_name:
                    self.setCurrentIndex(i)
                    return
            # If not found, default to empty
            self.setCurrentIndex(0)

    def currentSlot(self):
        """Get current slot name (or None for empty)"""
        return self.itemData(self.currentIndex())


class SlotsWidget(QWidget):
    """Widget with two slot comboboxes for asset cards."""

    slotsChanged = Signal(object)  # Emits list or None

    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        self.slot1_combo = SlotComboBox()
        self.slot2_combo = SlotComboBox()

        self.slot1_combo.currentIndexChanged.connect(self._on_changed)
        self.slot2_combo.currentIndexChanged.connect(self._on_changed)

        layout.addWidget(QLabel("Slot 1:"))
        layout.addWidget(self.slot1_combo)
        layout.addWidget(QLabel("Slot 2:"))
        layout.addWidget(self.slot2_combo)
        layout.addStretch()

        self.setLayout(layout)

    def _on_changed(self):
        """Emit the current slots value"""
        self.slotsChanged.emit(self.get_slots())

    def get_slots(self):
        """Get slots as list (reversed order for rendering) or None if both empty"""
        slot1 = self.slot1_combo.currentSlot()
        slot2 = self.slot2_combo.currentSlot()

        if not slot1 and not slot2:
            return None
        # Return as [right, left] due to rendering order, filtering out None
        result = [s for s in [slot2, slot1] if s]
        return result if result else None

    def set_slots(self, slots):
        """Set slots from list (reversed order) or None"""
        # Block signals to prevent triggering changes during load
        self.slot1_combo.blockSignals(True)
        self.slot2_combo.blockSignals(True)
        try:
            if not slots:
                self.slot1_combo.setCurrentSlot(None)
                self.slot2_combo.setCurrentSlot(None)
            elif len(slots) == 1:
                self.slot1_combo.setCurrentSlot(slots[0])
                self.slot2_combo.setCurrentSlot(None)
            else:
                # slots is [right, left], so reverse when setting
                self.slot1_combo.setCurrentSlot(slots[1] if len(slots) > 1 else None)
                self.slot2_combo.setCurrentSlot(slots[0])
        finally:
            self.slot1_combo.blockSignals(False)
            self.slot2_combo.blockSignals(False)


class FaceEditor(QWidget):
    """Base class for all face editors"""

    # Signal emitted when type changes and editor should be swapped
    type_changed = Signal(object)  # Emits the face object

    def __init__(self, face, parent=None):
        super().__init__(parent)
        self.face = face
        self.fields = {}
        self.updating = False  # Prevent recursive updates

        # Main layout
        self.main_layout = QVBoxLayout()
        self.setLayout(self.main_layout)

        # Always add type selector first
        self.add_type_selector()

        # Then setup custom fields
        self.setup_ui()

        # Load data
        self.load_data()

    def add_type_selector(self):
        """Add type combobox - shown in all editors"""
        type_layout = QHBoxLayout()
        type_label = QLabel("Type")
        type_label.setMinimumWidth(110)

        self.type_combo = NoScrollComboBox()
        self.type_combo.setEditable(True)
        self.type_combo.setInsertPolicy(QComboBox.NoInsert)
        self.type_combo.addItems(ALL_CARD_TYPES)
        self.type_combo.addItem("--- fullart variants ---")
        self.type_combo.addItems(FULLART_CARD_TYPES)

        # Add autocomplete
        completer = QCompleter(ALL_CARD_TYPES+FULLART_CARD_TYPES)
        completer.setCaseSensitivity(Qt.CaseInsensitive)
        completer.setCompletionMode(QCompleter.PopupCompletion)
        self.type_combo.setCompleter(completer)

        # Connect to handle changes - use signals that fire on "commit" not every keystroke
        # activated: fires when user selects from dropdown
        # lineEdit().editingFinished: fires on Enter or focus loss
        self.type_combo.activated.connect(self._on_type_committed)
        self.type_combo.lineEdit().editingFinished.connect(self._on_type_committed)

        self.fields['type'] = self.type_combo

        type_layout.addWidget(type_label)
        type_layout.addWidget(self.type_combo)
        type_widget = QWidget()
        type_widget.setLayout(type_layout)
        self.main_layout.addWidget(type_widget)

    def _on_type_committed(self, _=None):
        """Handle type field commit (Enter, focus loss, or dropdown selection)"""
        if self.updating:
            return

        value = self.type_combo.currentText()
        old_type = self.face.get('type')
        if value and value != old_type:
            self.face.set('type', value)
            # Emit signal to trigger editor swap
            self.type_changed.emit(self.face)

    def setup_ui(self):
        """Override in subclasses to add custom fields"""
        pass

    def load_data(self):
        """Load data from face into fields"""
        self.updating = True
        for field_name, widget in self.fields.items():
            value = self.face.get(field_name, '')
            self.set_widget_value(widget, value)
        self.updating = False

    def set_widget_value(self, widget, value):
        """Set widget value based on type"""
        if isinstance(widget, QLineEdit):
            if isinstance(value, list):
                widget.setText(', '.join(str(v) for v in value))
            else:
                widget.setText(str(value) if value else '')
        elif isinstance(widget, QTextEdit):
            widget.setPlainText(str(value) if value else '')
        elif isinstance(widget, QComboBox):
            widget.setCurrentText(str(value) if value else '')

    def get_widget_value(self, widget):
        """Get widget value based on type"""
        if isinstance(widget, QLineEdit):
            return widget.text().strip()
        elif isinstance(widget, QTextEdit):
            return widget.toPlainText().strip()
        elif isinstance(widget, QComboBox):
            return widget.currentText().strip()
        return ''

    # Fields that require type conversion
    INTEGER_FIELDS = {'illustration_pan_x', 'illustration_pan_y'}
    FLOAT_FIELDS = {'illustration_scale'}
    LIST_FIELDS = {'classes'}  # Fields stored as lists but displayed as comma-separated

    def on_field_changed(self, field_name):
        """Handle field change"""
        if self.updating:
            return

        widget = self.fields.get(field_name)
        if not widget:
            return

        value = self.get_widget_value(widget)

        # Convert value to appropriate type
        if value:
            if field_name in self.INTEGER_FIELDS:
                try:
                    value = int(value)
                except ValueError:
                    return  # Invalid integer, don't update
            elif field_name in self.FLOAT_FIELDS:
                try:
                    value = float(value)
                except ValueError:
                    return  # Invalid float, don't update
            elif field_name in self.LIST_FIELDS:
                # Convert comma-separated string to list
                value = [v.strip() for v in value.split(',') if v.strip()]
            self.face.set(field_name, value)
        else:
            self.face.set(field_name, None)

        # Emit data_changed signal on parent CardEditor if it exists
        parent = self.parent()
        while parent:
            if hasattr(parent, 'data_changed'):
                parent.data_changed.emit()
                break
            parent = parent.parent()

    def add_labeled_line(self, label, field_name):
        """Helper to add a labeled line edit"""
        widget = editors.LabeledLineEdit(label)
        widget.input.textChanged.connect(lambda: self.on_field_changed(field_name))
        self.fields[field_name] = widget.input
        self.main_layout.addWidget(widget)
        return widget

    def add_trait_field(self, label="Traits", field_name="traits"):
        """Helper to add a trait field with autocomplete"""
        widget = editors.LabeledTraitEdit(label)
        widget.input.textChanged.connect(lambda: self.on_field_changed(field_name))
        self.fields[field_name] = widget.input
        self.main_layout.addWidget(widget)
        return widget

    def add_labeled_text(self, label, field_name, use_arkham=False):
        """Helper to add a labeled text edit"""
        widget = editors.LabeledTextEdit(label, use_arkham_editor=use_arkham)
        widget.input.textChanged.connect(lambda: self.on_field_changed(field_name))
        self.fields[field_name] = widget.input
        self.main_layout.addWidget(widget)
        return widget

    def add_copyright_collection_row(self):
        """Add copyright and collection fields side by side at the bottom"""
        row_widget = QWidget()
        row_layout = QHBoxLayout()
        row_layout.setContentsMargins(0, 0, 0, 0)

        # Copyright field
        copyright_input = editors.LabeledLineEdit("Copyright")
        copyright_input.input.textChanged.connect(lambda: self.on_field_changed('copyright'))
        self.fields['copyright'] = copyright_input.input
        row_layout.addWidget(copyright_input)

        # Collection field
        collection_input = editors.LabeledLineEdit("Collection")
        collection_input.input.textChanged.connect(lambda: self.on_field_changed('collection'))
        self.fields['collection'] = collection_input.input
        row_layout.addWidget(collection_input)

        row_widget.setLayout(row_layout)
        self.main_layout.addWidget(row_widget)

    def add_illustration_widget(self):
        """Add illustration widget"""
        # Determine which side this face is on
        face_side = 'front' if self.face == self.face.card.front else 'back'
        illustration = editors.IllustrationWidget(face_side=face_side)
        self.illustration_widget = illustration

        # Add fields
        self.fields['illustration'] = illustration.path_input.input
        self.fields['illustration_pan_y'] = illustration.pan_y_input.input
        self.fields['illustration_pan_x'] = illustration.pan_x_input.input
        self.fields['illustration_scale'] = illustration.scale_input.input
        self.fields['illustrator'] = illustration.artist_input.input

        # Connect signals
        illustration.path_input.input.textChanged.connect(lambda: self.on_field_changed('illustration'))
        illustration.pan_y_input.input.textChanged.connect(lambda: self.on_field_changed('illustration_pan_y'))
        illustration.pan_x_input.input.textChanged.connect(lambda: self.on_field_changed('illustration_pan_x'))
        illustration.scale_input.input.textChanged.connect(lambda: self.on_field_changed('illustration_scale'))
        illustration.artist_input.input.textChanged.connect(lambda: self.on_field_changed('illustrator'))

        self.main_layout.addWidget(illustration)
        return illustration


class BaseEditor(FaceEditor):
    """Basic editor with just type field"""

    def setup_ui(self):
        """Type is already added by base class"""
        self.main_layout.addStretch()


class AssetEditor(FaceEditor):
    """Editor for asset cards"""

    def setup_ui(self):
        # Basic fields
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Subtitle", "subtitle")

        # Traits with autocomplete
        self.add_trait_field()

        # Grid for compact fields
        grid_widget = QWidget()
        grid_layout = QFormLayout()

        # Classes
        classes_input = editors.LabeledLineEdit("Classes")
        classes_input.input.setPlaceholderText("Comma-separated")
        classes_input.textChanged.connect(lambda: self.on_field_changed('classes'))
        self.fields['classes'] = classes_input.input
        grid_layout.addRow(classes_input)

        # Cost
        cost_input = editors.LabeledLineEdit("Cost")
        cost_input.input.textChanged.connect(lambda: self.on_field_changed('cost'))
        self.fields['cost'] = cost_input.input
        grid_layout.addRow(cost_input)

        # Level
        level_combo = QComboBox()
        level_combo.addItems(['None', '0', '1', '2', '3', '4', '5', 'Custom'])
        level_combo.currentTextChanged.connect(lambda: self.on_field_changed('level'))
        self.fields['level'] = level_combo
        grid_layout.addRow("Level", level_combo)

        grid_widget.setLayout(grid_layout)
        self.main_layout.addWidget(grid_widget)

        # Icons widget (separate from grid)
        self.icons_widget = editors.IconsWidget()
        self.icons_widget.iconsChanged.connect(self.on_icons_changed)
        self.main_layout.addWidget(self.icons_widget)

        # Continue with more grid fields
        grid_widget2 = QWidget()
        grid_layout2 = QFormLayout()

        # Health
        health_input = editors.LabeledLineEdit("Health")
        health_input.input.textChanged.connect(lambda: self.on_field_changed('health'))
        self.fields['health'] = health_input.input
        grid_layout2.addRow(health_input)

        # Sanity
        sanity_input = editors.LabeledLineEdit("Sanity")
        sanity_input.input.textChanged.connect(lambda: self.on_field_changed('sanity'))
        self.fields['sanity'] = sanity_input.input
        grid_layout2.addRow(sanity_input)

        grid_widget2.setLayout(grid_layout2)
        self.main_layout.addWidget(grid_widget2)

        # Slots widget (separate from grid for better layout)
        self.slots_widget = SlotsWidget()
        self.slots_widget.slotsChanged.connect(self.on_slots_changed)
        self.main_layout.addWidget(self.slots_widget)

        # Text fields
        self.add_labeled_text("Text", "text", use_arkham=True)
        self.add_labeled_text("Flavor", "flavor_text")

        # Illustration
        self.add_illustration_widget()

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()

    def load_data(self):
        """Load data including icons and slots widgets"""
        super().load_data()
        # Load icons separately
        icons_value = self.face.get('icons', '')
        self.icons_widget.set_icons_string(icons_value)
        # Load slots separately
        slots_value = self.face.get('slots')
        self.slots_widget.set_slots(slots_value)

    def on_icons_changed(self, icons_str):
        """Handle icons widget change"""
        if self.updating:
            return
        if icons_str:
            self.face.set('icons', icons_str)
        else:
            self.face.set('icons', None)

    def on_slots_changed(self, slots):
        """Handle slots widget change"""
        if self.updating:
            return
        self.face.set('slots', slots)


class EventEditor(FaceEditor):
    """Editor for event cards"""

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Subtitle", "subtitle")

        # Traits with autocomplete
        self.add_trait_field()

        # Grid for compact fields
        grid_widget = QWidget()
        grid_layout = QFormLayout()

        classes_input = editors.LabeledLineEdit("Classes")
        classes_input.input.textChanged.connect(lambda: self.on_field_changed('classes'))
        self.fields['classes'] = classes_input.input
        grid_layout.addRow(classes_input)

        cost_input = editors.LabeledLineEdit("Cost")
        cost_input.input.textChanged.connect(lambda: self.on_field_changed('cost'))
        self.fields['cost'] = cost_input.input
        grid_layout.addRow(cost_input)

        level_combo = QComboBox()
        level_combo.addItems(['None', '0', '1', '2', '3', '4', '5'])
        level_combo.currentTextChanged.connect(lambda: self.on_field_changed('level'))
        self.fields['level'] = level_combo
        grid_layout.addRow("Level", level_combo)

        grid_widget.setLayout(grid_layout)
        self.main_layout.addWidget(grid_widget)

        # Icons widget (separate from grid)
        self.icons_widget = editors.IconsWidget()
        self.icons_widget.iconsChanged.connect(self.on_icons_changed)
        self.main_layout.addWidget(self.icons_widget)

        self.add_labeled_text("Text", "text", use_arkham=True)
        self.add_labeled_text("Flavor", "flavor_text")
        self.add_illustration_widget()

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()

    def load_data(self):
        """Load data including icons widget"""
        super().load_data()
        # Load icons separately
        icons_value = self.face.get('icons', '')
        self.icons_widget.set_icons_string(icons_value)

    def on_icons_changed(self, icons_str):
        """Handle icons widget change"""
        if self.updating:
            return
        if icons_str:
            self.face.set('icons', icons_str)
        else:
            self.face.set('icons', None)


# Skill editor is similar to event
SkillEditor = EventEditor


class EnemyEditor(FaceEditor):
    """Editor for enemy cards"""

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Subtitle", "subtitle")

        # Traits with autocomplete
        self.add_trait_field()

        grid_widget = QWidget()
        grid_layout = QFormLayout()

        for field, label in [
            ("classes", "Classes"),
            ("attack", "Attack"),
            ("health", "Health"),
            ("evade", "Evade"),
            ("damage", "Damage"),
            ("horror", "Horror"),
            ("victory", "Victory"),
        ]:
            widget = editors.LabeledLineEdit(label)
            self.fields[field] = widget.input
            # Use a factory function to create proper closure
            def make_callback(field_name):
                return lambda: self.on_field_changed(field_name)
            widget.input.textChanged.connect(make_callback(field))
            grid_layout.addRow(widget)

        grid_widget.setLayout(grid_layout)
        self.main_layout.addWidget(grid_widget)

        self.add_labeled_text("Text", "text", use_arkham=True)
        self.add_labeled_text("Flavor", "flavor_text")
        self.add_illustration_widget()

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()


class TreacheryEditor(FaceEditor):
    """Editor for treachery cards"""

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Subtitle", "subtitle")
        self.add_trait_field()
        self.add_labeled_line("Classes", "classes")
        self.add_labeled_line("Victory", "victory")

        self.add_labeled_text("Text", "text", use_arkham=True)
        self.add_labeled_text("Flavor", "flavor_text")
        self.add_illustration_widget()

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()


class IconComboBox(QComboBox):
    """ComboBox that displays icons for connection symbols"""

    CONNECTION_SYMBOLS = [
        'None',
        'circle', 'circle_alt',
        'clover',
        'cross', 'cross_alt',
        'diamond', 'diamond_alt',
        'equals', 'equals_alt',
        'heart', 'heart_alt',
        'hourglass', 'hourglass_alt',
        'moon', 'moon_alt',
        'spade',
        'square', 'square_alt',
        'squiggle', 'squiggle_alt',
        'star', 'star_alt',
        't', 't_alt',
        'tear',
        'triangle', 'triangle_alt',
    ]

    def __init__(self, parent=None):
        super().__init__(parent)

        # Set fixed narrow width
        self.setFixedWidth(55)
        self.setIconSize(QSize(28, 28))

        # Add all symbols with icons
        from shoggoth.files import overlay_dir
        from PySide6.QtGui import QIcon, QPixmap

        for symbol in self.CONNECTION_SYMBOLS:
            if symbol == 'None':
                # Empty option - show dash
                self.addItem('—', userData=None)
            else:
                icon_path = overlay_dir / f"location_hi_{symbol}.png"
                if icon_path.exists():
                    icon = QIcon(str(icon_path))
                    self.addItem(icon, '', userData=symbol)
                else:
                    # Fallback to text if icon missing
                    self.addItem(symbol[:1].upper(), userData=symbol)

    def setCurrentSymbol(self, symbol):
        """Set current selection by symbol name"""
        if not symbol or symbol == 'None':
            self.setCurrentIndex(0)
        else:
            for i in range(self.count()):
                if self.itemData(i) == symbol:
                    self.setCurrentIndex(i)
                    break

    def currentSymbol(self):
        """Get current symbol name"""
        return self.itemData(self.currentIndex())


class LocationEditor(FaceEditor):
    """Editor for location cards"""

    # Available connection symbols
    CONNECTION_SYMBOLS = [
        'None',
        'circle', 'circle_alt',
        'clover',
        'cross', 'cross_alt',
        'diamond', 'diamond_alt',
        'equals', 'equals_alt',
        'heart', 'heart_alt',
        'hourglass', 'hourglass_alt',
        'moon', 'moon_alt',
        'spade',
        'square', 'square_alt',
        'squiggle', 'squiggle_alt',
        'star', 'star_alt',
        't', 't_alt',
        'tear',
        'triangle', 'triangle_alt',
    ]

    def setup_ui(self):
        # Connection symbol (this location's own symbol)
        conn_layout = QHBoxLayout()
        conn_label = QLabel("Connection Symbol")
        conn_label.setMinimumWidth(110)

        self.connection_combo = IconComboBox()
        self.connection_combo.currentIndexChanged.connect(lambda: self.on_connection_changed())
        self.fields['connection'] = self.connection_combo

        conn_layout.addWidget(conn_label)
        conn_layout.addWidget(self.connection_combo)
        conn_layout.addStretch()  # Push to left
        conn_widget = QWidget()
        conn_widget.setLayout(conn_layout)
        self.main_layout.addWidget(conn_widget)

        # Basic fields
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Subtitle", "subtitle")

        self.add_labeled_line("Shroud", "shroud")
        self.add_trait_field()
        self.add_labeled_line("Clues", "clues")
        self.add_labeled_line("Victory", "victory")

        # Connections section - single row of icon dropdowns
        connections_row = QHBoxLayout()
        connections_label = QLabel("Connections")
        connections_label.setMinimumWidth(110)
        connections_row.addWidget(connections_label)

        # Store connection combos for easy access
        self.connection_combos = []

        for i in range(6):
            combo = IconComboBox()
            combo.currentIndexChanged.connect(lambda: self.on_connections_changed())
            self.connection_combos.append(combo)
            connections_row.addWidget(combo)

        connections_row.addStretch()  # Push to left
        connections_widget = QWidget()
        connections_widget.setLayout(connections_row)
        self.main_layout.addWidget(connections_widget)

        # Text fields
        self.add_labeled_text("Text", "text", use_arkham=True)
        self.add_labeled_text("Flavor", "flavor_text")
        self.add_illustration_widget()

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()

    def on_connection_changed(self):
        """Handle connection symbol change"""
        if self.updating:
            return

        value = self.connection_combo.currentSymbol()
        self.face.set('connection', value)

    def on_connections_changed(self):
        """Handle connections list change"""
        if self.updating:
            return

        # Collect all connection values
        connections = []
        for combo in self.connection_combos:
            value = combo.currentSymbol()
            if value:
                connections.append(value)

        # Store as list or None if empty
        if connections:
            self.face.set('connections', connections)
        else:
            self.face.set('connections', None)

    def load_data(self):
        """Load data from face into fields - override to handle connections specially"""
        self.updating = True

        # Load regular fields
        for field_name, widget in self.fields.items():
            if field_name == 'connection':
                # Handle connection symbol
                value = self.face.get(field_name, '')
                self.connection_combo.setCurrentSymbol(value)
            elif field_name != 'connections':
                # Regular fields
                value = self.face.get(field_name, '')
                self.set_widget_value(widget, value)

        # Load connections list
        connections = self.face.get('connections', [])
        if not connections:
            connections = []
        elif isinstance(connections, str):
            # Handle comma-separated string format
            connections = [c.strip() for c in connections.split(',') if c.strip()]

        # Set combo boxes
        for i, combo in enumerate(self.connection_combos):
            if i < len(connections):
                combo.setCurrentSymbol(connections[i])
            else:
                combo.setCurrentSymbol(None)

        self.updating = False


class JsonEditor(FaceEditor):
    """Raw JSON editor for unknown types or manual editing"""

    def setup_ui(self):
        # Add a header explaining this is raw JSON mode
        header = QLabel("<b>Raw JSON Editor</b>")
        header.setStyleSheet("font-size: 14pt; padding: 5px;")
        self.main_layout.addWidget(header)

        info = QLabel("Edit the card face data directly as JSON. Changes update automatically when valid.")
        info.setWordWrap(True)
        info.setStyleSheet("color: #666; padding: 5px;")
        self.main_layout.addWidget(info)

        # Create the JSON editor with syntax highlighting
        self.json_editor = editors.ArkhamTextEdit()
        self.json_editor.setPlaceholderText("{\n  \"type\": \"asset\",\n  \"name\": \"Card Name\"\n}")

        # Enable JSON syntax highlighting
        try:
            from pygments.lexers import JsonLexer
            self.json_editor.set_lexer(JsonLexer())
        except:
            pass  # Syntax highlighting optional

        self.json_editor.textChanged.connect(self.on_json_changed)
        self.main_layout.addWidget(self.json_editor)

        # Validation status
        self.status_label = QLabel("")
        self.status_label.setStyleSheet("padding: 5px;")
        self.main_layout.addWidget(self.status_label)

        # Format button
        button_row = QHBoxLayout()

        format_btn = QPushButton("Format JSON")
        format_btn.clicked.connect(self.format_json)
        button_row.addWidget(format_btn)

        validate_btn = QPushButton("Validate")
        validate_btn.clicked.connect(self.validate_json)
        button_row.addWidget(validate_btn)

        button_row.addStretch()

        button_widget = QWidget()
        button_widget.setLayout(button_row)
        self.main_layout.addWidget(button_widget)

    def format_json(self):
        """Format the JSON text"""
        try:
            text = self.json_editor.toPlainText()
            if text.strip():
                data = json.loads(text)
                formatted = json.dumps(data, indent=2)
                self.json_editor.setPlainText(formatted)
                self.status_label.setText("✓ JSON formatted")
                self.status_label.setStyleSheet("color: green; padding: 5px;")
        except json.JSONDecodeError as e:
            self.status_label.setText(f"✗ Invalid JSON: {e}")
            self.status_label.setStyleSheet("color: red; padding: 5px;")

    def validate_json(self):
        """Validate the JSON"""
        try:
            text = self.json_editor.toPlainText()
            if text.strip():
                json.loads(text)
                self.status_label.setText("✓ Valid JSON")
                self.status_label.setStyleSheet("color: green; padding: 5px;")
            else:
                self.status_label.setText("⚠ Empty JSON")
                self.status_label.setStyleSheet("color: orange; padding: 5px;")
        except json.JSONDecodeError as e:
            self.status_label.setText(f"✗ Invalid JSON: {e}")
            self.status_label.setStyleSheet("color: red; padding: 5px;")

    def load_data(self):
        """Load face data as JSON"""
        super().load_data()  # load type field
        self.updating = True
        try:
            json_text = json.dumps(self.face.data, indent=2)
            self.json_editor.setPlainText(json_text)
            self.status_label.setText("✓ JSON loaded")
            self.status_label.setStyleSheet("color: green; padding: 5px;")
        except Exception as e:
            self.json_editor.setPlainText(f"# Error loading JSON: {e}")
            self.status_label.setText(f"✗ Load error: {e}")
            self.status_label.setStyleSheet("color: red; padding: 5px;")
        self.updating = False

    def on_json_changed(self):
        """Save JSON back to face"""
        if self.updating:
            return

        try:
            text = self.json_editor.toPlainText()
            if text.strip():
                data = json.loads(text)
                # Update face data
                self.face.data = data
                # Emit type_changed if type changed (to potentially switch editor)
                if 'type' in data and data['type'] != self.face.get('type'):
                    self.type_changed.emit()
                self.status_label.setText("✓ Saved")
                self.status_label.setStyleSheet("color: green; padding: 5px;")
        except json.JSONDecodeError as e:
            self.status_label.setText(f"✗ Invalid JSON: {str(e)[:50]}...")
            self.status_label.setStyleSheet("color: red; padding: 5px;")



class ActEditor(FaceEditor):
    """Editor for act cards (front side)"""

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Index", "index")
        self.add_labeled_line("Clues", "clues")

        self.add_labeled_text("Text", "text", use_arkham=True)
        self.add_labeled_text("Flavor", "flavor_text")
        self.add_illustration_widget()

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()


class ActBackEditor(FaceEditor):
    """Editor for act cards (back side)"""

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Index", "index")

        self.add_labeled_text("Text", "text", use_arkham=True)
        self.add_labeled_text("Flavor", "flavor_text")

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()


class AgendaEditor(FaceEditor):
    """Editor for agenda cards (front side)"""

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Index", "index")
        self.add_labeled_line("Doom", "doom")

        self.add_labeled_text("Text", "text", use_arkham=True)
        self.add_labeled_text("Flavor", "flavor_text")
        self.add_illustration_widget()

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()


class AgendaBackEditor(FaceEditor):
    """Editor for agenda cards (back side)"""

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Index", "index")

        self.add_labeled_text("Text", "text", use_arkham=True)
        self.add_labeled_text("Flavor", "flavor_text")

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()


class InvestigatorEditor(FaceEditor):
    """Editor for investigator cards (front side)"""

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Subtitle", "subtitle")
        self.add_trait_field()

        # Classes
        classes_input = editors.LabeledLineEdit("Classes")
        classes_input.input.textChanged.connect(lambda: self.on_field_changed('classes'))
        self.fields['classes'] = classes_input.input
        self.main_layout.addWidget(classes_input)

        # Stats in a grid
        box_widget = QWidget()
        box_layout = QHBoxLayout()

        for field, label in [
            ("willpower", "Willpower"),
            ("intellect", "Intellect"),
            ("combat", "Combat"),
            ("agility", "Agility"),
        ]:
            widget = editors.LabeledLineEdit(label)
            self.fields[field] = widget.input

            def make_callback(field_name):
                return lambda: self.on_field_changed(field_name)
            widget.input.textChanged.connect(make_callback(field))
            box_layout.addWidget(widget)

        box_widget.setLayout(box_layout)
        self.main_layout.addWidget(box_widget)

        # Stats in a grid
        box_widget = QWidget()
        box_layout = QHBoxLayout()
        for field, label in [
            ("health", "Health"),
            ("sanity", "Sanity"),
        ]:
            widget = editors.LabeledLineEdit(label)
            self.fields[field] = widget.input

            def make_callback(field_name):
                return lambda: self.on_field_changed(field_name)
            widget.input.textChanged.connect(make_callback(field))
            box_layout.addWidget(widget)

        box_widget.setLayout(box_layout)
        self.main_layout.addWidget(box_widget)

        # Text fields
        self.add_labeled_text("Text", "text", use_arkham=True)
        self.add_labeled_text("Flavor", "flavor_text")

        # Illustration
        self.add_illustration_widget()

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()


class InvestigatorBackEditor(FaceEditor):
    """Editor for investigator cards (back side) with deck building entries"""

    NUM_ENTRIES = 8

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Subtitle", "subtitle")
        self.add_trait_field()

        # Classes
        classes_input = editors.LabeledLineEdit("Classes")
        classes_input.input.setPlaceholderText("Comma-separated")
        classes_input.input.textChanged.connect(lambda: self.on_field_changed('classes'))
        self.fields['classes'] = classes_input.input
        self.main_layout.addWidget(classes_input)

        # Deck building entries section
        entries_group = QGroupBox("Deck Building Options")
        entries_layout = QVBoxLayout()
        entries_layout.setSpacing(15)
        entries_layout.setContentsMargins(6, 6, 6, 6)

        self.entry_widgets = []  # Store (header_input, text_input) pairs

        for i in range(self.NUM_ENTRIES):
            entry_widget = QWidget()
            entry_layout = QVBoxLayout()
            entry_layout.setContentsMargins(0, 0, 0, 0)

            # Header input (narrow)
            header_input = QLineEdit()
            header_input.setPlaceholderText(f"Header {i+1}")
            header_input.textChanged.connect(self.on_entries_changed)
            entry_layout.addWidget(header_input)

            # Value input (multi-line text field)
            value_input = QPlainTextEdit()
            value_input.setPlaceholderText(f"Value {i+1}")
            value_input.setFixedHeight(54)  # ~3 lines
            value_input.textChanged.connect(self.on_entries_changed)
            entry_layout.addWidget(value_input)

            entry_widget.setLayout(entry_layout)
            entries_layout.addWidget(entry_widget)

            self.entry_widgets.append((header_input, value_input))

        entries_group.setLayout(entries_layout)
        self.main_layout.addWidget(entries_group)

        # Flavor text
        self.add_labeled_text("Flavor", "flavor_text")

        # Illustration
        self.add_illustration_widget()

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()

    def load_data(self):
        """Load data from face into fields, including special entries handling"""
        self.updating = True

        # Load regular fields
        for field_name, widget in self.fields.items():
            value = self.face.get(field_name, '')
            self.set_widget_value(widget, value)

        # Load entries
        entries = self.face.get('entries', [])
        if not entries:
            entries = []

        for i, (header_input, value_input) in enumerate(self.entry_widgets):
            if i < len(entries) and isinstance(entries[i], list) and len(entries[i]) >= 2:
                header_input.setText(str(entries[i][0]) if entries[i][0] else '')
                # QPlainTextEdit uses setPlainText instead of setText
                value_input.setPlainText(str(entries[i][1]) if entries[i][1] else '')
            else:
                header_input.setText('')
                value_input.setPlainText('')

        self.updating = False

    def on_entries_changed(self):
        """Handle changes to entry fields"""
        if self.updating:
            return

        # Collect all entries
        entries = []
        text_parts = []

        for header_input, value_input in self.entry_widgets:
            header = header_input.text().strip()
            # QPlainTextEdit uses toPlainText() instead of text()
            value = value_input.toPlainText().strip()

            if header or value:
                entries.append([header, value])
                # Only add to text if BOTH header and value have content
                if header and value:
                    text_parts.append(f"<b>{header}</b> {value}")

        # Save entries list
        if entries:
            self.face.set('entries', entries)
        else:
            self.face.set('entries', None)

        # Save combined text field
        if text_parts:
            combined_text = "\n".join(text_parts)
            self.face.set('text', combined_text)
        else:
            self.face.set('text', None)

        # Emit data_changed signal
        parent = self.parent()
        while parent:
            if hasattr(parent, 'data_changed'):
                parent.data_changed.emit()
                break
            parent = parent.parent()


# Create aliases for similar editors
LocationBackEditor = LocationEditor


class ChaosEditor(FaceEditor):
    """Editor for chaos bag reference cards"""

    NUM_ENTRIES = 10

    def setup_ui(self):
        self.add_labeled_line("Difficulty", "difficulty")

        # Entries section
        entries_group = QGroupBox("Chaos Bag Entries")
        entries_layout = QVBoxLayout()
        entries_layout.setSpacing(8)
        entries_layout.setContentsMargins(6, 6, 6, 6)

        self.entry_widgets = []  # Store (token_input, text_input) pairs

        for i in range(self.NUM_ENTRIES):
            entry_widget = QWidget()
            entry_layout = QHBoxLayout()
            entry_layout.setContentsMargins(0, 0, 0, 0)

            # Token input (comma-separated list of tokens)
            token_input = QLineEdit()
            token_input.setPlaceholderText(f"Tokens (e.g. -1, -2)")
            token_input.setMaximumWidth(150)
            token_input.textChanged.connect(self.on_entries_changed)
            entry_layout.addWidget(token_input)

            # Text input
            text_input = QLineEdit()
            text_input.setPlaceholderText(f"Effect text")
            text_input.textChanged.connect(self.on_entries_changed)
            entry_layout.addWidget(text_input)

            entry_widget.setLayout(entry_layout)
            entries_layout.addWidget(entry_widget)

            self.entry_widgets.append((token_input, text_input))

        entries_group.setLayout(entries_layout)
        self.main_layout.addWidget(entries_group)

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()

    def load_data(self):
        """Load data from face into fields"""
        self.updating = True

        # Load regular fields
        for field_name, widget in self.fields.items():
            value = self.face.get(field_name, '')
            self.set_widget_value(widget, value)

        # Load entries
        entries = self.face.get('entries', [])
        if not entries:
            entries = []

        for i, (token_input, text_input) in enumerate(self.entry_widgets):
            if i < len(entries) and isinstance(entries[i], dict):
                # Token is a list, join with comma
                tokens = entries[i].get('token', [])
                if isinstance(tokens, list):
                    token_input.setText(', '.join(str(t) for t in tokens))
                else:
                    token_input.setText(str(tokens) if tokens else '')
                text_input.setText(str(entries[i].get('text', '')))
            else:
                token_input.setText('')
                text_input.setText('')

        self.updating = False

    def on_entries_changed(self):
        """Handle changes to entry fields"""
        if self.updating:
            return

        entries = []
        for token_input, text_input in self.entry_widgets:
            token_str = token_input.text().strip()
            text = text_input.text().strip()

            if token_str or text:
                # Parse tokens as comma-separated list
                tokens = [t.strip() for t in token_str.split(',') if t.strip()]
                entries.append({'token': tokens, 'text': text})

        if entries:
            self.face.set('entries', entries)
        else:
            self.face.set('entries', None)

        # Emit data_changed signal
        parent = self.parent()
        while parent:
            if hasattr(parent, 'data_changed'):
                parent.data_changed.emit()
                break
            parent = parent.parent()


class CustomizableEditor(FaceEditor):
    """Editor for customizable cards with upgrade options"""

    NUM_ENTRIES = 12

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_text("Text", "text", use_arkham=True)

        # Entries section (customization options)
        entries_group = QGroupBox("Customization Options")
        entries_layout = QVBoxLayout()
        entries_layout.setSpacing(6)
        entries_layout.setContentsMargins(6, 6, 6, 6)

        self.entry_widgets = []  # Store (cost_input, name_input, text_input) tuples

        for i in range(self.NUM_ENTRIES):
            entry_widget = QWidget()
            entry_layout = QHBoxLayout()
            entry_layout.setContentsMargins(0, 0, 0, 0)
            entry_layout.setSpacing(4)

            # Cost input (small integer field)
            cost_input = QSpinBox()
            cost_input.setRange(0, 10)
            cost_input.setFixedWidth(50)
            cost_input.setToolTip("XP Cost")
            cost_input.valueChanged.connect(self.on_entries_changed)
            entry_layout.addWidget(cost_input)

            # Name input
            name_input = QLineEdit()
            name_input.setPlaceholderText(f"Option name")
            name_input.setMaximumWidth(150)
            name_input.textChanged.connect(self.on_entries_changed)
            entry_layout.addWidget(name_input)

            # Text input
            text_input = QLineEdit()
            text_input.setPlaceholderText(f"Option effect")
            text_input.textChanged.connect(self.on_entries_changed)
            entry_layout.addWidget(text_input)

            entry_widget.setLayout(entry_layout)
            entries_layout.addWidget(entry_widget)

            self.entry_widgets.append((cost_input, name_input, text_input))

        entries_group.setLayout(entries_layout)
        self.main_layout.addWidget(entries_group)

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()

    def load_data(self):
        """Load data from face into fields"""
        self.updating = True

        # Load regular fields
        for field_name, widget in self.fields.items():
            value = self.face.get(field_name, '')
            self.set_widget_value(widget, value)

        # Load entries
        entries = self.face.get('entries', [])
        if not entries:
            entries = []

        for i, (cost_input, name_input, text_input) in enumerate(self.entry_widgets):
            if i < len(entries) and isinstance(entries[i], list) and len(entries[i]) >= 3:
                cost_input.setValue(int(entries[i][0]) if entries[i][0] else 0)
                name_input.setText(str(entries[i][1]) if entries[i][1] else '')
                text_input.setText(str(entries[i][2]) if entries[i][2] else '')
            else:
                cost_input.setValue(0)
                name_input.setText('')
                text_input.setText('')

        self.updating = False

    def on_entries_changed(self):
        """Handle changes to entry fields"""
        if self.updating:
            return

        entries = []
        for cost_input, name_input, text_input in self.entry_widgets:
            cost = cost_input.value()
            name = name_input.text().strip()
            text = text_input.text().strip()

            if name or text or cost > 0:
                entries.append([cost, name, text])

        if entries:
            self.face.set('entries', entries)
        else:
            self.face.set('entries', None)

        # Emit data_changed signal
        parent = self.parent()
        while parent:
            if hasattr(parent, 'data_changed'):
                parent.data_changed.emit()
                break
            parent = parent.parent()


class StoryEditor(FaceEditor):
    """Editor for story cards"""

    def setup_ui(self):
        self.add_labeled_line("Name", "name")
        self.add_labeled_line("Classes", "classes")

        self.add_labeled_text("Text", "text", use_arkham=True)

        # Copyright and collection
        self.add_copyright_collection_row()

        self.main_layout.addStretch()


# Mapping of editor types to classes
EDITOR_MAPPING = {
    'player': BaseEditor,
    'customizable_back': BaseEditor,
    'encounter': BaseEditor,
    'asset': AssetEditor,
    'event': EventEditor,
    'skill': SkillEditor,
    'investigator': InvestigatorEditor,
    'investigator_back': InvestigatorBackEditor,
    'location': LocationEditor,
    'location_back': LocationBackEditor,
    'treachery': TreacheryEditor,
    'enemy': EnemyEditor,
    'act': ActEditor,
    'act_back': ActBackEditor,
    'agenda': AgendaEditor,
    'agenda_back': AgendaBackEditor,
    'chaos': ChaosEditor,
    'customizable': CustomizableEditor,
    'story': StoryEditor,
}


def get_editor_for_face(face, parent=None):
    """Get the appropriate editor for a face"""
    editor_type = face.get_editor()
    editor_class = EDITOR_MAPPING.get(editor_type, JsonEditor)
    return editor_class(face, parent)


import jpype
import jpype.imports
from shoggoth.files import asset_dir

def run_conversion(java_path, jar_path, project_path, output_path):
    if not jar_path:
        raise Exception("Path to Strange Eons jar file not set. Press F1 and change the settings. This needs to be the .jar file - other versions won't work.")

    # start the JVM to make further imports available
    jpype.startJVM(
        java_path or jpype.getDefaultJVMPath(),
        f'-javaagent:{jar_path}',
        classpath=[f'{jar_path}'],
    )

    # import the JVM generated imports
    from shoggoth import strange_eons_parser
    from ca.cgjennings.apps.arkham import StrangeEons
    import javax
    import java
    from java.io import File


    # start SE in headless mode
    # This is using the keep-alive script to prevent it from shutting down
    # while we run our python stuff
    StrangeEons.main([
        '--run',
        str(asset_dir / 'js' / 'KeepAlive.js')
    ])

    # all script interactions must happen in a non-main thread
    @jpype.JImplements(java.lang.Runnable)
    class Launch:
        @jpype.JOverride
        def run(self):
            # for readability, we have our function elsewhere
            strange_eons_parser.run_import(project_path, output_path)

    # now run the class in a new thread
    javax.swing.SwingUtilities.invokeAndWait(Launch())

    # and now that we're done, we terminate the JVM
    # this will forcefully kill SE, despite the script
    # preventing it from closing.
    print('Done running strange_eons.py')
    return
